#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import contextlib
import copy
import logging
from typing import List

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects import edna_test_data
from mxcubecore.HardwareObjects.EDNACharacterisation import EDNACharacterisation
from mxcubecore.HardwareObjects.XSDataMXCuBEv1_4 import (
    XSDataInputMXCuBE,
    XSDataResultMXCuBE,
)
from mxcubecore.model import queue_model_objects as qmo

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3"


class EDNACharacterisationMockup(EDNACharacterisation):
    def __init__(self, name) -> None:
        super(EDNACharacterisationMockup, self).__init__(name)

    def input_from_params(self, data_collection, char_params) -> XSDataInputMXCuBE:
        return XSDataInputMXCuBE.parseString(self.edna_default_input)

    def characterise(self, edna_input) -> XSDataResultMXCuBE:
        return XSDataResultMXCuBE.parseString(edna_test_data.EDNA_RESULT_DATA)

    def is_running(self) -> bool:
        return False

    def get_default_characterisation_parameters(self) -> qmo.CharacterisationParameters:
        return super(
            EDNACharacterisationMockup, self
        ).get_default_characterisation_parameters()

    def dc_from_output(  # noqa: C901, PLR0915
        self, edna_result, reference_image_collection
    ) -> List[qmo.DataCollection]:
        data_collections = []

        crystal = copy.deepcopy(reference_image_collection.crystal)
        ref_proc_params = reference_image_collection.processing_parameters
        processing_parameters = copy.deepcopy(ref_proc_params)

        try:
            char_results = edna_result.getCharacterisationResult()
            edna_strategy = char_results.getStrategyResult()
            collection_plan = edna_strategy.getCollectionPlan()[0]
            wedges = collection_plan.getCollectionStrategy().getSubWedge()
        except Exception:
            logging.getLogger("HWR").exception("")
        else:
            try:
                resolution = (
                    collection_plan.getStrategySummary().getResolution().getValue()
                )
                resolution = round(resolution, 3)
            except AttributeError:
                resolution = None

            try:
                transmission = (
                    collection_plan.getStrategySummary().getAttenuation().getValue()
                )
                transmission = round(transmission, 2)
            except AttributeError:
                transmission = None

            try:
                screening_id = edna_result.getScreeningId().getValue()
            except AttributeError:
                screening_id = None

            for i in range(len(wedges)):
                wedge = wedges[i]
                exp_condition = wedge.getExperimentalCondition()
                goniostat = exp_condition.getGoniostat()
                beam = exp_condition.getBeam()

                acq = qmo.Acquisition()
                acq.acquisition_parameters = (
                    HWR.beamline.get_default_acquisition_parameters()
                )
                acquisition_parameters = acq.acquisition_parameters

                acquisition_parameters.centred_position = (
                    reference_image_collection.acquisitions[
                        0
                    ].acquisition_parameters.centred_position
                )

                acq.path_template = HWR.beamline.get_default_path_template()

                # Use the same path template as the reference_collection
                # and update the members the needs to be changed. Keeping
                # the directories of the reference collection.
                ref_pt = reference_image_collection.acquisitions[0].path_template

                acq.path_template = copy.deepcopy(ref_pt)
                acq.path_template.directory = "/".join(
                    ref_pt.directory.split("/")[0:-2]
                )
                acq.path_template.wedge_prefix = "w" + str(i + 1)
                acq.path_template.reference_image_prefix = str()

                if resolution:
                    acquisition_parameters.resolution = resolution

                if transmission:
                    acquisition_parameters.transmission = transmission

                if screening_id:
                    acquisition_parameters.screening_id = screening_id

                with contextlib.suppress(AttributeError):
                    acquisition_parameters.osc_start = (
                        goniostat.getRotationAxisStart().getValue()
                    )

                with contextlib.suppress(AttributeError):
                    acquisition_parameters.osc_end = (
                        goniostat.getRotationAxisEnd().getValue()
                    )

                with contextlib.suppress(AttributeError):
                    acquisition_parameters.osc_range = (
                        goniostat.getOscillationWidth().getValue()
                    )

                try:
                    num_images = int(
                        abs(
                            acquisition_parameters.osc_end
                            - acquisition_parameters.osc_start
                        )
                        / acquisition_parameters.osc_range
                    )

                    acquisition_parameters.first_image = 1
                    acquisition_parameters.num_images = num_images
                    acq.path_template.num_files = num_images
                    acq.path_template.start_num = 1

                except AttributeError:
                    logging.getLogger("HWR").exception("")

                with contextlib.suppress(AttributeError):
                    acquisition_parameters.transmission = (
                        beam.getTransmission().getValue()
                    )

                with contextlib.suppress(AttributeError):
                    acquisition_parameters.energy = round(
                        (123984.0 / beam.getWavelength().getValue()) / 10000.0, 4
                    )

                with contextlib.suppress(AttributeError):
                    acquisition_parameters.exp_time = beam.getExposureTime().getValue()

                dc = qmo.DataCollection([acq], crystal, processing_parameters)
                data_collections.append(dc)

        return data_collections



import os
import sys
import time

from xaloc import XalocJob
from XSDataCommon import (
    XSDataFile,
    XSDataString,
)
from XSDataMXCuBEv1_3 import XSDataResultMXCuBE

from mxcubecore.HardwareObjects.EDNACharacterisation import EDNACharacterisation

sys.path.append("/beamlines/bl13/controls/devel/pycharm/ALBAClusterClient")


root = os.environ["POST_PROCESSING_SCRIPTS_ROOT"]

sls_script = os.path.join(root, "edna-mx/strategy/edna-mx.strategy.sl")


class ALBADataAnalysis(EDNACharacterisation):
    def init(self):
        EDNACharacterisation.init(self)

    def prepare_input(self, edna_input):
        # used for strategy calculation (characterization) using data analysis cluster
        # ALBA specific

        firstImage = None

        for dataSet in edna_input.getDataSet():
            for imageFile in dataSet.imageFile:
                if imageFile.getPath() is None:
                    continue
                firstImage = imageFile.path.value
                break

        listImageName = os.path.basename(firstImage).split("_")
        prefix = "_".join(listImageName[:-2])
        run_number = listImageName[-2]
        i = 1

        if hasattr(edna_input, "process_directory"):
            edna_directory = os.path.join(
                edna_input.process_directory,
                "characterisation_%s_run%s_%d" % (prefix, run_number, i),
            )
            while os.path.exists(edna_directory):
                i += 1
                edna_directory = os.path.join(
                    edna_input.process_directory,
                    "characterisation_%s_run%s_%d" % (prefix, run_number, i),
                )
            os.makedirs(edna_directory)
        else:
            raise RuntimeError("No process directory specified in edna_input")

        edna_input.process_directory = edna_directory

        output_dir = XSDataFile()
        path = XSDataString()
        path.set_value(edna_directory)
        output_dir.setPath(path)
        edna_input.setOutputFileDirectory(output_dir)

    def run_edna(self, input_file, results_file, edna_directory):
        return self.run(input_file, results_file, edna_directory)

    def run(self, *args):
        input_file, results_file, edna_directory = args

        jobname = os.path.basename(os.path.dirname(edna_directory))

        self.log.debug("  XalocJob submiting ")
        self.log.debug("      job_name: %s" % jobname)
        self.log.debug("      sls_script: %s, " % sls_script)
        self.log.debug("      input file: %s" % input_file)
        self.log.debug("      results file: %s" % results_file)
        self.log.debug("      edna directory: %s" % edna_directory)

        self.job = XalocJob(
            "edna-strategy", jobname, sls_script, input_file, edna_directory
        )
        self.job.submit()

        self.log.debug("  XalocJob submitted %s" % self.job.id)

        self.edna_directory = os.path.dirname(input_file)
        self.input_file = os.path.basename(input_file)
        # self.results_file = self.fix_path(results_file)
        self.results_file = results_file
        self.log.debug("      self.results file: %s" % self.results_file)

        state = self.wait_done()

        if state == "COMPLETED":
            self.log.debug("EDNA Job completed")
            time.sleep(0.5)
            result = self.get_result()
        else:
            self.log.debug(
                "EDNA Job finished without success / state was %s" % (self.job.state)
            )
            result = ""

        return result

    def fix_path(self, path):
        outpath = path.replace("PROCESS_DATA", "PROCESS_DATA/RESULTS")
        # dirname = os.path.dirname(path)
        # basename = os.path.basename(path)
        # outpath = os.path.join(dirname,'RESULTS',basename)
        return outpath

    def wait_done(self):
        self.log.debug("Polling for Job state")
        time.sleep(0.5)
        self.log.debug("Polling for Job state 2")

        try:
            state = self.job.state
            self.log.debug("Job / is %s" % str(state))
        except Exception:
            self.log.debug("Polling for Job state 3. exception happened")
            self.log.exception("")

        while state in ["RUNNING", "PENDING"]:
            self.log.debug("Job / is %s" % state)
            time.sleep(0.5)
            state = self.job.state

        self.log.debug("Returning")
        self.log.debug("Returning %s" % str(state))
        return state

    def get_result(self):
        jobstatus = self.job.status

        # outname = self.input_file.replace("Input", "Output")
        # outfile = os.path.join( self.edna_directory, outname)

        self.log.debug("Job / state is COMPLETED")
        self.log.debug("  job status dump: %s" % jobstatus)
        self.log.debug("  looking for file: %s" % self.results_file)

        if os.path.exists(self.results_file):
            # job_output = open(outfile).read()
            # self.log.debug("     EDNA results file found. loading it")
            # open(self.results_file, "w").write(job_output)
            self.log.debug("     EDNA results file found 2")
            result = XSDataResultMXCuBE.parseFile(self.results_file)
            self.log.debug("     EDNA results file found 3")
            self.log.debug(
                "EDNA Result loaded from file / result is=%s" % str(type(result))
            )
        else:
            self.log.debug(
                "EDNA Job finished without success / cannot find output file "
            )
            result = ""

        return result


def test_hwo(hwo):
    ofile = "/tmp/edna/edna_result"
    odir = "/tmp/edna"
    test_input_file = "/beamlines/bl13/projects/cycle2018-I/2018012551-bcalisto/mx2018012551/DATA/20180131/PROCESS_DATA/characterisation_ref-Thrombin-TB-TTI1_A_run1_1/EDNAInput_2004391.xml"
    result = hwo.run_edna(test_input_file, ofile, odir)
    print(result)



import binascii
import copy
import logging
import os
import subprocess
import time
from typing import List

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractCharacterisation import (
    AbstractCharacterisation,
)
from mxcubecore.HardwareObjects.SecureXMLRpcRequestHandler import (
    SecureXMLRpcRequestHandler,
)
from mxcubecore.HardwareObjects.XSDataCommon import (
    XSDataAngle,
    XSDataBoolean,
    XSDataDouble,
    XSDataFlux,
    XSDataImage,
    XSDataInteger,
    XSDataLength,
    XSDataSize,
    XSDataString,
    XSDataTime,
    XSDataWavelength,
)
from mxcubecore.HardwareObjects.XSDataMXCuBEv1_4 import (
    XSDataInputMXCuBE,
    XSDataMXCuBEDataSet,
    XSDataResultMXCuBE,
)
from mxcubecore.model import queue_model_enumerables as qme
from mxcubecore.model import queue_model_objects as qmo

# from edna_test_data import EDNA_DEFAULT_INPUT
# from edna_test_data import EDNA_TEST_DATA


class EDNACharacterisation(AbstractCharacterisation):
    def __init__(self, name) -> None:
        super(EDNACharacterisation, self).__init__(name)

        self.result = None
        self.edna_default_file = None
        self.start_edna_command = None

    def init(self) -> None:
        self.start_edna_command = self.get_property("edna_command")
        self.edna_default_file = self.get_property("edna_default_file")

        fp = HWR.get_hardware_repository().find_in_repository(self.edna_default_file)

        if fp is None:
            fp = self.edna_default_file

            if not os.path.exists(fp):
                raise ValueError("File %s not found in repository" % fp)

        with open(fp, "r") as f:
            self.edna_default_input = "".join(f.readlines())

    def _modify_strategy_option(self, diff_plan, strategy_option) -> None:
        """Method for modifying the diffraction plan 'strategyOption' entry"""
        if diff_plan.getStrategyOption() is None:
            new_strategy_option = strategy_option
        else:
            new_strategy_option = (
                diff_plan.getStrategyOption().getValue() + " " + strategy_option
            )

        diff_plan.setStrategyOption(XSDataString(new_strategy_option))

    def _run_edna(
        self, input_file, results_file, process_directory
    ) -> XSDataResultMXCuBE:
        """Starts EDNA"""
        msg = "Starting EDNA characterisation using xml file %s" % input_file
        logging.getLogger("queue_exec").info(msg)
        self.characterisationResult = None
        args = (self.start_edna_command, input_file, results_file, process_directory)
        # subprocess.call("%s %s %s %s" % args, shell=True)
        p = subprocess.Popen("%s %s %s %s --verbose --debug" % args, shell=True)

        do_continue = True
        self.result = None
        start_time = time.time()
        TIME_OUT = 120
        while do_continue:
            if self.characterisationResult is not None:
                logging.getLogger("queue_exec").info(
                    "Received characterisation results via XMLRPC"
                )
                self.result = XSDataResultMXCuBE.parseString(
                    self.characterisationResult
                )
                do_continue = False
            elif p.poll() is not None:
                do_continue = False
            elif time.time() - start_time > TIME_OUT:
                do_continue = False
            if do_continue:
                logging.getLogger("queue_exec").info(
                    "Waiting for characterisation results..."
                )
                time.sleep(1)

        if self.result is None and os.path.exists(results_file):
            self.result = XSDataResultMXCuBE.parseFile(results_file)

        return self.result

    def get_html_report(self, edna_result) -> str:
        """
        Returns the path to the html result report generated by the characterisation software.

        Args:
            output (EDNAResult) EDNAResult object
        """
        html_report = None

        try:
            html_report = str(edna_result.getHtmlPage().getPath().getValue())
        except AttributeError:
            pass

        return html_report

    def input_from_params(self, data_collection, char_params) -> XSDataInputMXCuBE:
        edna_input = XSDataInputMXCuBE.parseString(self.edna_default_input)

        if data_collection.id:
            edna_input.setDataCollectionId(XSDataInteger(data_collection.id))

        # Beam object
        beam = edna_input.getExperimentalCondition().getBeam()

        try:
            transmission = HWR.beamline.transmission.get_value()
            beam.setTransmission(XSDataDouble(transmission))
        except AttributeError:
            self.log.exception("EDNACharacterisation. transmission not saved ")

        try:
            wavelength = HWR.beamline.energy.get_wavelength()
            beam.setWavelength(XSDataWavelength(wavelength))
        except AttributeError:
            pass

        try:
            beam.setFlux(XSDataFlux(HWR.beamline.flux.get_value()))
        except AttributeError:
            pass

        try:
            min_exp_time = HWR.beamline.detector.get_exposure_time_limits()[0]
            beam.setMinExposureTimePerImage(XSDataTime(min_exp_time))
        except AttributeError:
            logging.getLogger("HWR").exception("")

        try:
            beamsize = HWR.beamline.beam.beam_info_hwobj.get_beam_size()

            if None not in beamsize:
                beam.setSize(
                    XSDataSize(
                        x=XSDataLength(float(beamsize[0])),
                        y=XSDataLength(float(beamsize[1])),
                    )
                )
        except AttributeError:
            pass

        # Optimization parameters
        diff_plan = edna_input.getDiffractionPlan()

        aimed_i_sigma = XSDataDouble(char_params.aimed_i_sigma)
        aimed_completness = XSDataDouble(char_params.aimed_completness)
        aimed_multiplicity = XSDataDouble(char_params.aimed_multiplicity)
        aimed_resolution = XSDataDouble(char_params.aimed_resolution)

        complexity = char_params.strategy_complexity
        complexity = XSDataString(qme.STRATEGY_COMPLEXITY[complexity])

        permitted_phi_start = XSDataAngle(char_params.permitted_phi_start)
        _range = char_params.permitted_phi_end - char_params.permitted_phi_start
        rotation_range = XSDataAngle(_range)

        if char_params.aimed_i_sigma:
            diff_plan.setAimedIOverSigmaAtHighestResolution(aimed_i_sigma)

        if char_params.aimed_completness:
            diff_plan.setAimedCompleteness(aimed_completness)

        if char_params.use_aimed_multiplicity:
            diff_plan.setAimedMultiplicity(aimed_multiplicity)

        if char_params.use_aimed_resolution:
            diff_plan.setAimedResolution(aimed_resolution)

        diff_plan.setComplexity(complexity)
        diff_plan.setStrategyType(XSDataString(char_params.strategy_program))

        if char_params.use_permitted_rotation:
            diff_plan.setUserDefinedRotationStart(permitted_phi_start)
            diff_plan.setUserDefinedRotationRange(rotation_range)

        # Vertical crystal dimension
        sample = edna_input.getSample()
        sample.getSize().setY(XSDataLength(char_params.max_crystal_vdim))
        sample.getSize().setZ(XSDataLength(char_params.min_crystal_vdim))

        # Radiation damage model
        sample.setSusceptibility(XSDataDouble(char_params.rad_suscept))
        sample.setChemicalComposition(None)
        sample.setRadiationDamageModelBeta(XSDataDouble(char_params.beta / 1e6))
        sample.setRadiationDamageModelGamma(XSDataDouble(char_params.gamma / 1e6))

        diff_plan.setForcedSpaceGroup(XSDataString(char_params.space_group))

        # Characterisation type - Routine DC
        if char_params.use_min_dose:
            pass

        if char_params.use_min_time:
            time = XSDataTime(char_params.min_time)
            diff_plan.setMaxExposureTimePerDataCollection(time)

        # Account for radiation damage
        if char_params.induce_burn:
            self._modify_strategy_option(diff_plan, "-DamPar")

        # Characterisation type - SAD
        if char_params.opt_sad:
            if char_params.auto_res:
                diff_plan.setAnomalousData(XSDataBoolean(True))
            else:
                diff_plan.setAnomalousData(XSDataBoolean(False))
                self._modify_strategy_option(diff_plan, "-SAD yes")
                diff_plan.setAimedResolution(XSDataDouble(char_params.sad_res))
        else:
            diff_plan.setAnomalousData(XSDataBoolean(False))

        # Data set
        data_set = XSDataMXCuBEDataSet()
        acquisition_parameters = data_collection.acquisitions[0].acquisition_parameters
        path_template = data_collection.acquisitions[0].path_template
        path_str = os.path.join(
            path_template.directory, path_template.get_image_file_name()
        )
        characterisation_dir = path_template.xds_dir.replace(
            "/autoprocessing_", "/characterisation_"
        )
        os.makedirs(characterisation_dir, mode=0o755, exist_ok=True)
        for img_num in range(int(acquisition_parameters.num_images)):
            image_file = XSDataImage()
            path = XSDataString()
            path.value = path_str % (img_num + 1)
            image_file.path = path
            image_file.number = XSDataInteger(img_num + 1)
            data_set.addImageFile(image_file)

        edna_input.addDataSet(data_set)
        edna_input.process_directory = characterisation_dir
        return edna_input

    def characterise(self, edna_input) -> str:
        """
        Runs Characterisation and returns the results.

        Args:
            input (EDNAInput) EDNA input object
        """
        self.processing_done_event.set()
        self.prepare_input(edna_input)
        path = edna_input.process_directory

        # if there is no data collection id, the id will be a random number
        # this is to give a unique number to the EDNA input and result files;
        # something more clever might be done to give a more significant
        # name, if there is no dc id.
        try:
            dc_id = edna_input.getDataCollectionId().getValue()
        except Exception:
            dc_id = id(edna_input)

        token = self.generate_new_token()
        edna_input.token = XSDataString(token)

        if hasattr(edna_input, "process_directory"):
            edna_input_file = os.path.join(path, "EDNAInput_%s.xml" % dc_id)
            edna_input.exportToFile(edna_input_file)
            edna_results_file = os.path.join(path, "EDNAOutput_%s.xml" % dc_id)

            if not os.path.isdir(path):
                os.makedirs(path)
        else:
            raise RuntimeError("No process directory specified in edna_input")

        self.result = self._run_edna(edna_input_file, edna_results_file, path)

        self.processing_done_event.clear()
        return self.result

    def dc_from_output(
        self, edna_result, reference_image_collection
    ) -> List[qmo.DataCollection]:
        data_collections = []

        crystal = copy.deepcopy(reference_image_collection.crystal)
        ref_proc_params = reference_image_collection.processing_parameters
        processing_parameters = copy.deepcopy(ref_proc_params)

        try:
            char_results = edna_result.getCharacterisationResult()
            edna_strategy = char_results.getStrategyResult()
            collection_plan = edna_strategy.getCollectionPlan()[0]
            wedges = collection_plan.getCollectionStrategy().getSubWedge()
        except Exception:
            pass
        else:
            try:
                resolution = (
                    collection_plan.getStrategySummary().getResolution().getValue()
                )
                resolution = round(resolution, 3)
            except AttributeError:
                resolution = None

            try:
                transmission = (
                    collection_plan.getStrategySummary().getAttenuation().getValue()
                )
                transmission = round(transmission, 2)
            except AttributeError:
                transmission = None

            try:
                screening_id = edna_result.getScreeningId().getValue()
            except AttributeError:
                screening_id = None

            for i in range(0, len(wedges)):
                wedge = wedges[i]
                exp_condition = wedge.getExperimentalCondition()
                goniostat = exp_condition.getGoniostat()
                beam = exp_condition.getBeam()

                acq = qmo.Acquisition()
                acq.acquisition_parameters = (
                    HWR.beamline.get_default_acquisition_parameters()
                )
                acquisition_parameters = acq.acquisition_parameters

                acquisition_parameters.centred_position = (
                    reference_image_collection.acquisitions[
                        0
                    ].acquisition_parameters.centred_position
                )

                acq.path_template = HWR.beamline.get_default_path_template()

                # Use the same path template as the reference_collection
                # and update the members the needs to be changed. Keeping
                # the directories of the reference collection.
                ref_pt = reference_image_collection.acquisitions[0].path_template

                acq.path_template = copy.deepcopy(ref_pt)
                acq.path_template.directory = "/".join(
                    ref_pt.directory.split("/")[0:-2]
                )
                acq.path_template.wedge_prefix = "w" + str(i + 1)
                acq.path_template.reference_image_prefix = str()

                if resolution:
                    acquisition_parameters.resolution = resolution

                if transmission:
                    acquisition_parameters.transmission = transmission

                if screening_id:
                    acquisition_parameters.screening_id = screening_id

                try:
                    acquisition_parameters.osc_start = (
                        goniostat.getRotationAxisStart().getValue()
                    )
                except AttributeError:
                    pass

                try:
                    acquisition_parameters.osc_end = (
                        goniostat.getRotationAxisEnd().getValue()
                    )
                except AttributeError:
                    pass

                try:
                    acquisition_parameters.osc_range = (
                        goniostat.getOscillationWidth().getValue()
                    )
                except AttributeError:
                    pass

                try:
                    num_images = int(
                        abs(
                            acquisition_parameters.osc_end
                            - acquisition_parameters.osc_start
                        )
                        / acquisition_parameters.osc_range
                    )

                    acquisition_parameters.first_image = 1
                    acquisition_parameters.num_images = num_images
                    acq.path_template.num_files = num_images
                    acq.path_template.start_num = 1

                except AttributeError:
                    pass

                try:
                    acquisition_parameters.transmission = (
                        beam.getTransmission().getValue()
                    )
                except AttributeError:
                    pass

                try:
                    acquisition_parameters.energy = round(
                        (123984.0 / beam.getWavelength().getValue()) / 10000.0, 4
                    )
                except AttributeError:
                    pass

                try:
                    acquisition_parameters.exp_time = beam.getExposureTime().getValue()
                except AttributeError:
                    pass

                dc = qmo.DataCollection([acq], crystal, processing_parameters)
                data_collections.append(dc)

        return data_collections

    def get_default_characterisation_parameters(self) -> qmo.CharacterisationParameters:
        """
        Returns the default parameters
        """
        edna_input = XSDataInputMXCuBE.parseString(self.edna_default_input)
        diff_plan = edna_input.getDiffractionPlan()

        edna_sample = edna_input.getSample()
        char_params = qmo.CharacterisationParameters()
        char_params.experiment_type = qme.EXPERIMENT_TYPE.OSC

        # Optimisation parameters
        char_params.use_aimed_resolution = False
        try:
            char_params.aimed_resolution = diff_plan.getAimedResolution().getValue()
        except Exception:
            char_params.aimed_resolution = None

        char_params.use_aimed_multiplicity = False
        try:
            char_params.aimed_i_sigma = (
                diff_plan.getAimedIOverSigmaAtHighestResolution().getValue()
            )
            char_params.aimed_completness = diff_plan.getAimedCompleteness().getValue()
        except Exception:
            char_params.aimed_i_sigma = None
            char_params.aimed_completness = None

        char_params.strategy_complexity = 0
        char_params.induce_burn = False
        char_params.use_permitted_rotation = False
        char_params.permitted_phi_start = 0.0
        char_params.permitted_phi_end = 360
        char_params.low_res_pass_strat = False

        # Crystal
        char_params.max_crystal_vdim = edna_sample.getSize().getY().getValue()
        char_params.min_crystal_vdim = edna_sample.getSize().getZ().getValue()
        char_params.max_crystal_vphi = 90
        char_params.min_crystal_vphi = 0.0
        char_params.space_group = ""

        # Characterisation type
        char_params.use_min_dose = True
        char_params.use_min_time = False
        char_params.min_dose = 30.0
        char_params.min_time = 0.0
        char_params.account_rad_damage = True
        char_params.auto_res = True
        char_params.opt_sad = False
        char_params.sad_res = 0.5
        char_params.determine_rad_params = False
        char_params.burn_osc_start = 0.0
        char_params.burn_osc_interval = 3

        # Radiation damage model
        char_params.rad_suscept = edna_sample.getSusceptibility().getValue()
        char_params.beta = 1
        char_params.gamma = 0.06

        return char_params

    def generate_new_token(self) -> str:
        # See: https://wyattbaldwin.com/2014/01/09/generating-random-tokens-in-python/
        token = binascii.hexlify(os.urandom(5)).decode("utf-8")
        SecureXMLRpcRequestHandler.setReferenceToken(token)
        return token



from math import (
    asin,
    atan,
    sin,
    tan,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.LNLS.EPICS.EPICSMotor import EPICSMotor


class ResolutionVirtualMotor(EPICSMotor):
    """
    This class implements the resolution virtual motor.

    YAML Example
    ------------

    %YAML 1.2
    ---
    class: LNLS.LNLSResolution.ResolutionVirtualMotor
    epics:
    "MNC:B:PB04:CS1:m9":
        channels:
        "":
    "MX2:cam1:BeamX":
        channels:
            beam_x:
                suffix: ""
                polling_period: 200
    "MX2:cam1:BeamY":
        channels:
            beam_y:
                suffix: ""
                polling_period: 200
    "MX2:cam1:Wavelength":
        channels:
            wavelength_rbv:
                suffix: ""
                polling_period: 200
    configuration:
        tolerance: 0.01
        pixel_size_mm: 0.172
        n_pixels_x: 1475
        n_pixels_y: 1679
    """

    BEAM_X_RBV = "beam_x"
    BEAM_Y_RBV = "beam_y"

    def __init__(self, name: str):
        super().__init__(name)
        self.wavelength = HWR.beamline.get_object_by_role("wavelength")

    def init(self):
        self.pixel_size_mm = self.get_property("pixel_size_mm")
        self.n_pixels_x = self.get_property("n_pixels_x")
        self.n_pixels_y = self.get_property("n_pixels_y")
        self.dx = self.n_pixels_x * self.pixel_size_mm
        self.dy = self.n_pixels_y * self.pixel_size_mm
        super().init()
        self.connect(
            self.get_channel_object("wavelength_rbv"), "update", self.update_value
        )

    def get_limits(self):
        llm, hlm = super().get_limits()
        low_limit = self.distance_to_resolution(llm)
        high_limit = self.distance_to_resolution(hlm)
        return (low_limit, high_limit)

    def get_radius(self):
        beam_x = self.get_channel_value(self.BEAM_X_RBV) * self.pixel_size_mm
        radius_x = min(beam_x, self.dx - beam_x)
        beam_y = self.get_channel_value(self.BEAM_Y_RBV) * self.pixel_size_mm
        radius_y = min(beam_y, self.dy - beam_y)
        return min(radius_x, radius_y)

    def distance_to_resolution(self, distance):
        wavelength = self.wavelength.get_value()
        radius = self.get_radius()
        try:
            two_theta = atan(radius / distance)
            theta = two_theta / 2
            return wavelength / (2 * sin(theta))
        except ZeroDivisionError:
            self.print_log(level="error", msg="Distance cannot be zero")
            return None
        except TypeError as e:
            self.print_log(level="error", msg=f"Invalid numeric input: {e}")
            return None

    def resolution_to_distance(self, resolution):
        wavelength = self.wavelength.get_value()
        radius = self.get_radius()
        try:
            theta = asin(wavelength / (2 * resolution))
            two_theta = 2 * theta
            return radius / tan(two_theta)
        except ZeroDivisionError:
            self.print_log(level="error", msg="Resolution cannot be zero")
            return None
        except TypeError as e:
            self.print_log(level="error", msg=f"Invalid numeric input: {e}")
            return None

    def get_value(self):
        distance = super().get_value()
        return self.distance_to_resolution(distance)

    def update_value(self, value=None) -> None:
        value = self.get_value()
        super().update_value(value)

    def _set_value(self, value):
        resolution = value
        distance = self.resolution_to_distance(resolution)
        if distance is not None:
            super()._set_value(distance)



from mxcubecore.BaseHardwareObjects import HardwareObject

DETECTOR_DIAMETER = 424.0

NOTINITIALIZED, UNUSABLE, READY, MOVESTARTED, MOVING, ONLIMIT = (0, 1, 2, 3, 4, 5)


class PX1Resolution(HardwareObject):
    stateDict = {
        "UNKNOWN": 0,
        "ALARM": 1,
        "STANDBY": 2,
        "RUNNING": 4,
        "MOVING": 4,
        "FAULT": 1,
        "1": 1,
        "2": 2,
    }

    def _init(self):
        self._nominal_value = None
        self.currentDistance = None

        self.connect("equipmentReady", self.equipmentReady)
        self.connect("equipmentNotReady", self.equipmentNotReady)

        self.distance_chan = self.get_channel_object("distance")
        self.resolution_chan = self.get_channel_object("resolution")
        self.minimum_res_chan = self.get_channel_object("minimum_resolution")
        self.maximum_res_chan = self.get_channel_object("maximum_resolution")
        self.minimum_dist_chan = self.get_channel_object("minimum_distance")
        self.state_chan = self.get_channel_object("state")

        self.stop_command = self.get_command_object("stop")

        self.distance_chan.connect_signal("update", self.distanceChanged)
        self.resolution_chan.connect_signal("update", self.resolutionChanged)
        self.minimum_res_chan.connect_signal("update", self.minimumResolutionChanged)
        self.maximum_res_chan.connect_signal("update", self.maximumResolutionChanged)
        self.minimum_dist_chan.connect_signal("update", self.minimumDistanceChanged)
        self.state_chan.connect_signal("update", self.stateChanged)

        self.currentDistance = self.distance_chan.get_value()
        self._nominal_value = self.resolution_chan.get_value()

        return super()._init()

    def connect_notify(self, signal):
        if signal == "stateChanged":
            self.stateChanged()
        elif signal == "distanceChanged":
            self.distanceChanged()
        elif signal == "resolutionChanged":
            self.resolutionChanged()
        elif signal == "distanceLimitsChanged":
            self.minimumResolutionChanged()
        elif signal == "resolutionLimitsChanged":
            self.minimumResolutionChanged()

    def equipmentReady(self):
        self.emit("deviceReady")

    def equipmentNotReady(self):
        self.emit("deviceNotReady")

    def get_state(self, value=None):
        if value is None:
            value = self.state_chan.get_value()
        state_str = str(value)
        # return self.stateDict[state_str]
        return state_str

    def get_value(self):
        if self._nominal_value is None:
            self.recalculateResolution()
        return self._nominal_value

    def getDistance(self):
        if self._nominal_value is None:
            self.recalculateResolution()
        return self.currentDistance

    def minimumResolutionChanged(self, value=None):
        self.emit("resolutionLimitsChanged", (self.get_limits(),))

    def maximumResolutionChanged(self, value=None):
        self.emit("resolutionLimitsChanged", (self.get_limits(),))

    def minimumDistanceChanged(self, value=None):
        self.emit("distanceLimitsChanged", (self.getDistanceLimits(),))

    def stateChanged(self, state=None):
        self.emit("stateChanged", (self.get_state(state),))

    def distanceChanged(self, value=None):
        self.recalculateResolution()

    def resolutionChanged(self, value=None):
        self.recalculateResolution()

    def recalculateResolution(self):
        distance = self.distance_chan.get_value()
        resolution = self.resolution_chan.get_value()

        if resolution is None or distance is None:
            return

        if (self._nominal_value is not None) and abs(
            resolution - self._nominal_value
        ) > 0.001:
            self._nominal_value = resolution
            self.emit("resolutionChanged", (resolution,))

        if (self.currentDistance is not None) and abs(
            distance - self.currentDistance
        ) > 0.001:
            self.currentDistance = distance
            self.emit("distanceChanged", (distance,))

    def getDistanceLimits(self):
        chan_info = self.distance_chan.getInfo()

        high = float(chan_info.max_value)
        low = self.minimum_dist_chan.get_value()

        return [low, high]

    def get_limits(self):
        high = self.maximum_res_chan.get_value()
        low = self.minimum_res_chan.get_value()

        return (low, high)

    def moveResolution(self, res):
        self.resolution_chan.set_value(res)

    def moveDistance(self, dist):
        self.distance_chan.set_value(dist)

    def stop(self):
        try:
            self.stop_command()
        except Exception:
            self.log.err("%s: PX1Resolution.stop: error while trying to stop!", self.id)

    def re_emit_values(self):
        self.stateChanged()
        self.distanceChanged()
        self.resolutionChanged()
        self.minimumResolutionChanged()
        self.minimumResolutionChanged()

    move = moveResolution


def test_hwo(hwo):
    print("Distance [limits]", hwo.getDistance(), hwo.getDistanceLimits())
    print("Resolution [limits]", hwo.get_value(), hwo.get_limits())



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


import base64
import copy
import logging
import math
from ast import literal_eval
from functools import reduce
from io import BytesIO

import numpy as np
from gevent import GreenletExit
from PIL import Image

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects import sample_centring
from mxcubecore.HardwareObjects.abstract.AbstractSampleView import (
    AbstractSampleView,
    ShapeState,
)
from mxcubecore.model import queue_model_objects as qmo

__copyright__ = """by the MXCuBE collaboration """
__license__ = "LGPLv3+"


def combine_images(img1, img2):
    if img1.size != img2.size:
        raise ValueError("Images must be the same size")

    combined_img = Image.new("RGB", img1.size)

    pixels1 = img1.load()
    pixels2 = img2.load()
    combined_pixels = combined_img.load()

    width, height = img1.size
    for x in range(width):
        for y in range(height):
            pixel1 = pixels1[x, y]
            pixel2 = pixels2[x, y]

            if pixel2[0] <= 200 and pixel2[1] <= 60 and pixel2[2] <= 140:
                combined_pixels[x, y] = pixel1
            else:
                combined_pixels[x, y] = pixel2

    return combined_img


class SampleView(AbstractSampleView):
    """SampleView class"""

    def __init__(self, name):
        super().__init__(name)
        self.centring_motors = {}
        self.current_centring_procedure = None
        self.current_centring_method = None
        self.centring_status = {}
        self.rotation_reference = {}

    def init(self):
        super().init()

        centring_motor_roles = self.get_property("centring_motors", [])
        if isinstance(centring_motor_roles, str):
            centring_motor_roles = literal_eval(centring_motor_roles)
        # need to set the motor names for the centring points
        qmo.CentredPosition.DIFFRACTOMETER_MOTOR_NAMES = centring_motor_roles

        centring_ref_position = self.get_property("centring_reference_position", {})
        if isinstance(centring_ref_position, str):
            centring_ref_position = literal_eval(centring_ref_position)

        motor_directions = self.get_property("motor_directions", {})
        if isinstance(motor_directions, str):
            motor_directions = literal_eval(motor_directions)
        diffr = HWR.beamline.diffractometer

        for role in centring_motor_roles:
            if role in diffr.motors_hwobj_dict:
                motor_obj = diffr.motors_hwobj_dict[role]
                ref_position = None
                if role in centring_ref_position:
                    ref_position = centring_ref_position[role]
                direction = motor_directions.get(role, 1)
                self.centring_motors[role] = sample_centring.CentringMotor(
                    motor_obj, reference_position=ref_position, direction=direction
                )
                self.centring_motors[role].motor.connect(
                    "stateChanged", self._update_shape_positions
                )
        self._camera = self.get_object_by_role("camera")
        self._last_oav_image = None

        self.hide_grid_threshold = self.get_property("hide_grid_threshold", 5)
        self.centring_status = {"valid": False}
        rotation_reference = self.get_property("rotation_reference", {})
        if isinstance(rotation_reference, str):
            self.rotation_reference = literal_eval(rotation_reference)
        self.rotation_reference.update(
            {"motor": self.centring_motors.get(self.rotation_reference.get("name"))}
        )

    def _update_shape_positions(self, *args, **kwargs):
        for shape in self.get_shapes():
            shape.update_position(self.motor_positions_to_screen)

        self.emit("shapesChanged")

    def get_positions(self) -> dict[str, float]:
        """Get motor positions for the centring motors.

        Returns:
            Centring motor positions as {role: position}
        """
        motors_dict = {}
        for key, val in self.centring_motors.items():
            motors_dict.update({key: val.motor.get_value()})
        return motors_dict

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        """Get the motor positions form x,y pixel coordinates"""

        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        diffr = HWR.beamline.diffractometer
        pixels_per_mm_x, pixels_per_mm_y = diffr.get_pixels_per_mm()
        if not all([pixels_per_mm_x, pixels_per_mm_y]):
            return 0, 0

        # distance from the point to the beam
        dx = (x - beam_pos_x) / pixels_per_mm_x
        dy = (y - beam_pos_y) / pixels_per_mm_y

        motors_dict = self.get_positions()
        for key, val in motors_dict.items():
            motors_dict.update({key: self.centring_motors[key].direction * val})

        omega_angle = math.radians(motors_dict.get("omega", 0))
        rot_matrix = np.matrix(
            [
                [math.cos(omega_angle), -math.sin(omega_angle)],
                [math.sin(omega_angle), math.cos(omega_angle)],
            ]
        )
        inv_rot_matrix = np.array(rot_matrix.I)
        dsampx, dsampy = np.dot(np.array([0, dy]), inv_rot_matrix)

        chi_angle = math.radians(motors_dict.get("chi", 0))
        chi_rot = np.matrix(
            [
                [math.cos(chi_angle), -math.sin(chi_angle)],
                [math.sin(chi_angle), math.cos(chi_angle)],
            ]
        )
        sx, sy = np.dot(np.array([dsampx, dsampy]), np.array(chi_rot))

        sampx = -motors_dict.get("sampx") + sx
        sampy = motors_dict.get("sampy") + sy
        phiy = motors_dict.get("phiy") + dx

        return {
            "omega": motors_dict.get("omega"),
            "phiy": float(-phiy),
            "phiz": motors_dict.get("phiz"),
            "sampx": float(-sampx),
            "sampy": float(sampy),
        }

    def motor_positions_to_screen(
        self, positions_dict: dict[str, float]
    ) -> tuple[int, int]:
        """Get the x,y pixel value according to the calibration.

        Args:
            positions_dict: Dictionary {role: position}
        """
        if not positions_dict:
            raise RuntimeError("Unknown position")
        try:
            diffr = HWR.beamline.diffractometer
            p_x, p_y = diffr.get_pixels_per_mm()
            if None in (p_x, p_y):
                return 0, 0
            omega_angle = math.radians(-diffr.omega.get_value())
            sampx = positions_dict.get("sampx") - diffr.sampx.get_value()
            sampy = positions_dict.get("sampy") - diffr.sampy.get_value()
            phiy = -(positions_dict.get("phiy") - diffr.phiy.get_value())
            phiz = positions_dict.get("phiz") - diffr.phiz.get_value()

            rot_matrix = np.matrix(
                [
                    [math.cos(omega_angle), -math.sin(omega_angle)],
                    [math.sin(omega_angle), math.cos(omega_angle)],
                ]
            )
            inv_rot_matrix = np.array(rot_matrix.I)
            _, dy = np.dot(np.array([sampx, sampy]), inv_rot_matrix) * p_x

            chi_angle = math.radians(positions_dict.get("chi", 0))
            chi_rot = np.matrix(
                [
                    [math.cos(chi_angle), -math.sin(chi_angle)],
                    [math.sin(chi_angle), math.cos(chi_angle)],
                ]
            )
            sx, sy = np.dot(np.array([0, dy]), np.array(chi_rot))

            beam_position = HWR.beamline.beam.get_beam_position_on_screen()

            x = sx + (phiy * p_x) + beam_position[0]
            y = sy + (phiz * p_y) + beam_position[1]

        except AttributeError as err:
            raise NotImplementedError from err
        return x, y

    def start_manual_centring(self, nb_click: int = 3):
        """Do the manual centring procedure.

        Args:
           nb_click: Number of clicks.
        """
        if self.current_centring_procedure is not None:
            logging.getLogger("HWR").exception("Already centring")

        self.current_centring_method = "Manual"
        self.emit("centringStarted", ("Manual"))
        beam_pos = HWR.beamline.beam.get_beam_position_on_screen()
        diffr = HWR.beamline.diffractometer
        pixels_per_mm = diffr.get_pixels_per_mm()
        diffr.wait_status_ready(5)

        self.current_centring_procedure = sample_centring.start(
            self.centring_motors,
            pixels_per_mm[0],
            pixels_per_mm[1],
            beam_pos[0],
            beam_pos[1],
            chi_angle=0.0,
        )

        self.current_centring_procedure.link(self.manual_centring_done)

    def get_centring_status(self):
        return copy.deepcopy(self.centring_status)

    def image_clicked(self, x, y):
        logging.getLogger("user_level_log").info(
            f"Centring click at x:{int(x)}, y:{int(y)}"
        )
        sample_centring.user_click(x, y, wait=True)

    def manual_centring_done(self, manual_centring_procedure):
        try:
            motor_pos = manual_centring_procedure.get()
            if isinstance(motor_pos, GreenletExit):
                raise motor_pos
        except Exception:
            logging.exception("Could not complete manual centring")
            self.centring_failed()
        else:
            try:
                sample_centring.end()
            except Exception:
                logging.exception("Could not move to centred position")
                self.centring_failed()

            self.centring_done()

    def auto_centring_done(self, auto_centring_procedure):
        try:
            res = auto_centring_procedure.get()
        except Exception:
            logging.exception("Could not complete automatic centring")
            logging.getLogger("user_level_log").info("Automatic loop centring failed")
            self.centring_failed()
        else:
            if res is None:
                logging.error("Could not complete automatic centring")
                logging.getLogger("user_level_log").info(
                    "Automatic loop centring failed"
                )
                self.centring_failed()
            else:
                self.centring_done()
                self.accept_centring()

    def centring_done(self):
        """Execute if centring accepted."""
        self.centring_status = {"motors": {}, "method": self.current_centring_method}
        self.centring_status["motors"] = self.get_positions()
        HWR.beamline.diffractometer.save_centring_positions()

        self.centring_status["valid"] = True

        self.emit(
            "centringSuccessful",
            (self.current_centring_method, self.get_centring_status()),
        )
        self.current_centring_method = None
        self.current_centring_procedure = None

    def accept_centring(self):
        """Accept the current centred position."""
        self.centring_status["valid"] = True
        self.centring_status["accepted"] = True
        self.emit("centringAccepted", (True, self.get_centring_status()))
        logging.getLogger("user_level_log").info("Centring successful")

    def reject_centring(self):
        """Reject the current centred position."""
        if self.current_centring_procedure:
            self.current_centring_procedure.kill(block=True)
        self.centring_status["valid"] = False
        self.emit("centringAccepted", (False, self.get_centring_status()))
        logging.getLogger("user_level_log").info("Centring cancelled")

    def cancel_centring(self):
        """Cancel current centring procedure."""
        if self.current_centring_procedure:
            try:
                self.current_centring_procedure.kill(block=True)
            except Exception:
                logging.getLogger("HWR").exception(
                    "Problem aborting the centring method"
                )

            logging.getLogger("HWR").exception("Centring canceled")
        self.centring_failed()

    def centring_failed(self):
        """Execute if centring failed or canceled."""
        self.centring_status["valid"] = False
        self.emit(
            "centringFailed", (self.current_centring_method, self.get_centring_status())
        )
        self.current_centring_procedure = None
        self.current_centring_method = None

    def run_custom_auto_centring(self):
        """Run custom script for the automatic centring"""
        self.log.info("Using custom sample centring")
        self.current_centring_method = "Automatic"
        self.emit("centringStarted", ("Automatic"))
        diffr = HWR.beamline.diffractometer
        diffr.run_custom_script("sample_centering")
        diffr.wait_status_ready()
        self.centring_done()
        self.accept_centring()

    def start_auto_centring(self):
        """Start automatic centring procedure"""
        if self.current_centring_procedure is not None:
            logging.getLogger("HWR").exception("Already centring")

        diffr = HWR.beamline.diffractometer
        diffr.wait_status_ready(60)
        diffr.set_phase(diffr.get_phase_enum.CENTRE)

        if self.get_property("use_custom_auto_centring"):
            self.run_custom_auto_centring()
        else:
            beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()

            pixels_per_mm_x, pixels_per_mm_y = diffr.get_pixels_per_mm()

            self.current_centring_procedure = sample_centring.start_auto(
                self,
                self.centring_motors,
                pixels_per_mm_x,
                pixels_per_mm_y,
                beam_pos_x,
                beam_pos_y,
                chi_angle=0.0,
            )

            self.current_centring_method = "Automatic"
            self.emit("centringStarted", ("Automatic"))
            self.current_centring_procedure.link(self.auto_centring_done)

    def move_to_beam(self, x: float, y: float):
        """Move the sample to the x,y coordinates.
        Args:
            x: Pixels on x axis
            y: Pixels on y axis
        """
        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        diffr = HWR.beamline.diffractometer
        pixels_per_mm_x, pixels_per_mm_y = diffr.get_pixels_per_mm()
        if not all([pixels_per_mm_x, pixels_per_mm_y]):
            logging.getLogger("HWR").exception("Cannot move to beam")

        # here added the calculation for moving to the beam position
        dx = (x - beam_pos_x) / pixels_per_mm_x
        dy = (y - beam_pos_y) / pixels_per_mm_y

        diffr.wait_status_ready(5)
        motors_dict = self.get_positions()
        for key, val in motors_dict.items():
            motors_dict.update({key: self.centring_motors[key].direction * val})
        omega_angle = math.radians(motors_dict.get("omega", 0))

        rot_matrix = np.matrix(
            [
                [math.cos(omega_angle), -math.sin(omega_angle)],
                [math.sin(omega_angle), math.cos(omega_angle)],
            ]
        )
        inv_rot_matrix = np.array(rot_matrix.I)
        dsampx, dsampy = np.dot(np.array([0, dy]), inv_rot_matrix)

        chi_angle = math.radians(motors_dict.get("chi", 0))
        chi_rot = np.matrix(
            [
                [math.cos(chi_angle), -math.sin(chi_angle)],
                [math.sin(chi_angle), math.cos(chi_angle)],
            ]
        )

        sx, sy = np.dot(np.array([dsampx, dsampy]), np.array(chi_rot))

        sampx = -motors_dict.get("sampx") + sx
        sampy = motors_dict.get("sampy") + sy
        phiy = motors_dict.get("phiy") + dx

        self.centring_motors.get("sampx").set_value(-sampx)
        self.centring_motors.get("sampy").set_value(sampy)
        self.centring_motors.get("phiy").set_value(-phiy)
        diffr.save_centring_positions()

    def get_snapshot(
        self,
        overlay: str | None = None,
        bw: bool = False,
        return_as_array: bool = False,
    ) -> BytesIO:
        """Get snapshot(s)

        Args:
            overlay: Image data with shapes and other items to display
                          on the snapshot
            bw: return grayscale image
            return_as_array: return as np array if True, Default False

        Returns:
            (BytesIO) snapshot as bytes image
        """
        img = self.take_snapshot(overlay_data=overlay, bw=bw)

        if return_as_array:
            return np.array(img)

        buffered = BytesIO()
        img.save(buffered, format="JPEG")

        return buffered

    def take_acq_snapshot(self, image_path_list: list):
        """Take snapshot in the acquisition sequence.
        Args:
           image_path_list: List of file name(s) to save the snapshot(s}
                            (full path).
        """
        if len(image_path_list) > 0:
            diffr = HWR.beamline.diffractometer
            phase = diffr.get_phase_enum.CENTRE
            if diffr.get_phase() != phase:
                use_custom_snapshot_routine = (
                    self.get_property("custom_snapshot_script_dir") or False
                )

                if not use_custom_snapshot_routine:
                    diffr.set_phase(phase)

        for image_path in image_path_list:
            snapshot_index = image_path_list.index(image_path)
            msg = f"Taking {snapshot_index + 1} sample snapshot(s)"
            self.log.info(msg)

            self.save_snapshot(filename=image_path)
            # do not move 90 degrees if not needed
            if not diffr.in_plate_mode and snapshot_index < len(image_path_list) - 1:
                diffr.omega.set_value_relative(90, timeout=200)

    def save_snapshot(
        self,
        filename: str,
        overlay: str | None = None,
        bw: bool = False,
    ):
        """Save a snapshot to file.

        Args:
            filename: The filename.
            overlay: Image data with shapes and other items to display
                      on the snapshot
            bw): return grayscale image if true. Default False
        """
        img = self.take_snapshot(overlay_data=overlay, bw=bw)
        img.save(filename)

        self._last_oav_image = filename

    def take_snapshot(self, overlay_data: str | None = None, bw: bool = False):
        """Get snapshot with overlaid data.

        Args:
            overlay_data: base64 encoded image to lay over camera image
            bw: return grayscale image if True, Default False

        Returns:
            (Image) rgb or grayscale image
        """
        data, width, height = self.camera.get_last_image()

        img = Image.frombytes("RGB", (width, height), data)

        if overlay_data:
            overlay_data = base64.b64decode(overlay_data)
            overlay_image = Image.open(BytesIO(overlay_data))
            overlay_image = overlay_image.resize(
                (width, height), Image.Resampling.LANCZOS
            )
            img = combine_images(img, overlay_image.convert("RGB"))

        if bw:
            img.convert("1")

        return img

    def get_last_image_path(self):
        return self._last_oav_image

    def add_shape(self, shape):
        """Add the shape <shape> to the dictionary of handled shapes.

        Args:
            shape: Shape to add.
        """
        self.shapes[shape.id] = shape
        shape.shapes_hw_object = self

    def add_shape_from_mpos(
        self,
        mpos_list: list[float],
        screen_coord: tuple[int, int],
        t: str,
        state: ShapeState = "SAVED",
        user_state: ShapeState = "SAVED",
    ):
        """
        Adds a shape of type <t>, with motor positions from mpos_list and
        screen position screen_coord.

        Args:
            mpos_list: List of motor positions
            screen_coord: Screen coordinate for shape (x, y)
            t: Type str for shape, P (Point), L (Line), G (Grid)

        Returns:
            (Shape) Shape of type <t>
        """
        cls_dict = {"P": Point, "L": Line, "G": Grid, "2DP": TwoDPoint}
        _cls = cls_dict[t]
        shape = None

        if _cls:
            shape = _cls(mpos_list, screen_coord)
            # In case the shape is being recreated, we need to restore it's state.
            shape.state = state
            shape.user_state = user_state

            self.add_shape(shape)

        return shape

    def add_shape_from_refs(
        self, refs, t, state: ShapeState = "SAVED", user_state: ShapeState = "SAVED"
    ):
        """
        Adds a shape of type <t>, taking motor positions and screen positions
        from reference points in refs.

        Args:
            refs (list[str]): List of id's of the reference Points
            t (str): Type str for shape, P (Point), L (Line), G (Grid)

        Returns:
            (Shape): Shape of type <t>
        """
        mpos = [self.get_shape(refid).mpos() for refid in refs]
        spos_list = [self.get_shape(refid).screen_coord for refid in refs]
        spos = reduce((lambda x, y: tuple(x) + tuple(y)), spos_list, ())
        shape = self.add_shape_from_mpos(mpos, spos, t, state, user_state)
        shape.refs = refs

        return shape

    def delete_shape(self, sid):
        """
        Removes the shape with id <sid> from the list of handled shapes.

        Args:
            sid (str): The id of the shape to remove

        Returns:
            (Shape): The removed shape
        """
        shape = self.shapes.pop(sid, None)

        if shape:
            shape.shapes_hw_object = None

        return shape

    def select_shape(self, sid):
        """
        Select the shape <shape>.

        Args:
            sid (str): Id of the shape to select.
        """
        shape = self.shapes.get(sid, None)

        if shape:
            shape.select()

    def de_select_shape(self, sid):
        """
        De-select the shape with id <sid>.

        Args:
            sid (str): The id of the shape to de-select.
        """
        shape = self.shapes.get(sid, None)

        if shape:
            shape.de_select()

    def is_selected(self, sid):
        """
        Check if Shape with <sid> is selected.

        Returns:
            (Boolean) True if Shape with <sid> is selected False otherwise
        """
        shape = self.shapes.get(sid, None)
        return bool(shape and shape.is_selected())

    def get_selected_shapes(self):
        """
        Get all selected shapes.

        Returns:
           (list[Shape]) List of selected Shapes
        """
        return [s for s in self.shapes.values() if s.is_selected()]

    def de_select_all(self):
        """De select all shapes."""

        for shape in self.shapes.values():
            shape.de_select()

    def select_shape_with_cpos(self, cpos):
        """
        Selects shape with the assocaitaed centred position <cpos>

        Args:
            cpos (CenteredPosition)
        """
        return

    def clear_all(self):
        """
        Clear the shapes, remove all contents.
        """
        self._shapes = {}
        Grid.SHAPE_COUNT = 0
        Line.SHAPE_COUNT = 0
        Point.SHAPE_COUNT = 0

    def get_shapes(self) -> list:
        """Get all Shapes.

        Returns:
            (list[Shape]) All the shapes
        """
        return self.shapes.values()

    def get_points(self) -> list:
        """Get all Points currently handled.

        Returns:
            List[Point] - All currently handled points.
        """
        current_points = []

        for shape in self.get_shapes():
            if isinstance(shape, Point):
                current_points.append(shape)

        return current_points

    def get_lines(self) -> list:
        """Get all Lines currently handled.

        Returns:
            List[Line] - All currently handled lines.
        """
        lines = []

        for shape in self.get_shapes():
            if isinstance(shape, Line):
                lines.append(shape)

        return lines

    def get_grids(self) -> list:
        """Get all Grids currently handled.

        Returns:
            List[Grid] - All currently handled grids,
        """
        grid = []

        for shape in self.get_shapes():
            if isinstance(shape, Grid):
                grid.append(shape)

        return grid

    def get_shape(self, sid: str):
        """Get Shape with id <sid>.

        Args:
            sid: id of Shape to retrieve

        Returns:
            (Shape) All the shapes
        """
        return self.shapes.get(sid, None)

    # For backwards compatibility with old ShapeHisotry object
    # returns first of selected grids
    def get_grid(self):
        """
        Get the first of the selected grids, (the one that was selected first in
        a sequence of select operations)

        Returns:
            (dict): The first selected grid as a dictionary
        """
        grid = None

        for shape in self.get_shapes():
            if isinstance(shape, Grid):
                grid = shape.as_dict()
                break

        return grid

    def set_grid_data(self, sid: str, result_data, data_file_path: str):
        """
        Sets grid rsult data for a shape with the specified id.

        Args:
            sid: The id of the shape to set grid data for.
            result_data: The result data to set for the shape.
                         Either a base64 encoded string for PNG/image or a
                         dictionary for RGB (keys are cell number and value
                         RGBa list). Data is only updated if result is RGB based
            data_file_path: The path to the data file associated with the
                            result data.

        Raises:
            AttributeError: If no shape with the specified id exists.
        """

        shape = self.get_shape(sid)

        if shape:
            if shape.result and isinstance(shape.result, dict):
                # append data
                shape.result.update(result_data)
            else:
                shape.set_result(result_data)
                shape.result_data_path = data_file_path

            self.emit("newGridResult", shape)
        else:
            msg = f"Cant set result, no shape with id {sid}"
            raise AttributeError(msg)

    def get_grid_data(self, key):
        result = {}
        shape = self.get_shape(key)

        if shape:
            result = shape.get_result()

        return result

    def inc_used_for_collection(self, cpos):
        """
        Increase counter that keepts on collect made on this shape,
        shape with associated CenteredPosition cpos

        Args:
            cpos (CenteredPosition): CenteredPosition of shape
        """

    def set_rotation_axis_position(self, value: float):
        """Set the reference position for the rotation axis.

        value: the position
        """
        motor = self.rotation_reference.get("motor")
        if motor:
            self.log.info(f"Setting rotation axis ({motor.name}) position to {value}")
            self.centring_motors[motor.name].reference_position = value
            script_name = self.rotation_reference["script"]
            try:
                self.log.info("Setting MD Alignment reference position")
                HWR.beamline.diffractometer.run_script(f"{script_name}, {value}")
            except:
                self.log.exception("Setting Alignment reference position failed")
                raise


class Shape:
    """
    Base class for shapes.
    """

    SHAPE_COUNT = 0

    def __init__(self, mpos_list=None, screen_coord=(-1, -1)):
        Shape.SHAPE_COUNT += 1
        self.t = "S"
        self.id = ""
        self.cp_list = []
        self.name = ""
        self.state: ShapeState = "SAVED"
        # used to persist user preferences to show or hide particular shape.
        self.user_state: ShapeState = "SAVED"
        self.label = ""
        self.screen_coord = screen_coord
        self.selected = False
        self.refs = []
        self.shapes_hw_object = None
        mpos_list = mpos_list or []
        self.add_cp_from_mp(mpos_list)

    def get_centred_positions(self):
        """
        :returns: The centred position(s) associated with the shape.
        :rtype: List of CentredPosition objects.
        """
        return self.cp_list

    def get_centred_position(self):
        return self.get_centred_positions()[0]

    def select(self):
        self.selected = True

    def de_select(self):
        self.selected = False

    def is_selected(self):
        return self.selected

    def update_position(self, transform):
        spos_list = [transform(cp.as_dict()) for cp in self.cp_list]
        spos_list = tuple(pos for x in spos_list for pos in x)
        self.screen_coord = spos_list

    def add_cp_from_mp(self, mpos_list):
        for mp in mpos_list:
            self.cp_list.append(qmo.CentredPosition(mp))

    def set_id(self, id_num):
        self.id = f"{self.t}{id_num}"
        self.name = f"{self.label}-{id_num}"

    def move_to_mpos(self, mpos_list, screen_coord=None):
        self.cp_list = []
        self.add_cp_from_mp(mpos_list)

        if screen_coord:
            self.screen_coord = screen_coord

    def update_from_dict(self, shape_dict):
        # We do not allow id or result updates
        shape_dict.pop("id", None)
        shape_dict.pop("result", None)

        for key, value in shape_dict.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def as_dict(self):
        cpos_list = [x.as_dict() for x in self.cp_list]

        d = copy.deepcopy(vars(self))

        # Do not serialize Shapes HW Object
        d.pop("shapes_hw_object")

        # replace cpos_list with a list of motor positions
        d.pop("cp_list")
        d["motor_positions"] = str(cpos_list)

        return d


class Point(Shape):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        Shape.__init__(self, mpos_list, screen_coord)
        Point.SHAPE_COUNT += 1
        self.t = "P"
        self.label = "Point"
        self.set_id(Point.SHAPE_COUNT)

    def mpos(self):
        return self.cp_list[0].as_dict()

    def set_id(self, id_num):
        Shape.set_id(self, id_num)
        self.cp_list[0].index = self.name

    def as_dict(self):
        d = Shape.as_dict(self)
        # replace cpos_list with the motor positions
        d["motor_positions"] = self.cp_list[0].as_dict()
        return d


class TwoDPoint(Point):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        Point.__init__(self, mpos_list, screen_coord)
        self.t = "2DP"
        self.label = "2D-Point"
        self.set_id(Point.SHAPE_COUNT)


class Line(Shape):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        super().__init__(mpos_list, screen_coord)
        Line.SHAPE_COUNT += 1
        self.t = "L"
        self.label = "Line"
        self.set_id(Line.SHAPE_COUNT)

    def set_id(self, id_num):
        Shape.set_id(self, id_num)
        self.cp_list[0].index = self.name

    def get_points_index(self):
        if all(self.cp_list):
            return (self.cp_list[0].get_index(), self.cp_list[1].get_index())
        return (None, None)


class Grid(Shape):
    SHAPE_COUNT = 0

    def __init__(self, mpos_list, screen_coord):
        super().__init__(mpos_list, screen_coord)
        Grid.SHAPE_COUNT += 1
        self.t = "G"
        self.set_id(Grid.SHAPE_COUNT)

        self.width = -1
        self.height = -1
        self.cell_count_fun = "zig-zag"
        self.cell_h_space = -1
        self.cell_height = -1
        self.cell_v_space = -1
        self.cell_width = -1
        self.label = "Grid"
        self.num_cols = -1
        self.num_rows = -1
        self.selected = False
        # result is a base64 encoded string for PNG/image heatmap results
        # or a dictionary (for RGB number based results)
        self.result = None
        self.pixels_per_mm = [1, 1]
        self.beam_pos = [1, 1]
        self.beam_width = 0
        self.beam_height = 0
        self.hide_threshold = 5

        self.set_id(Grid.SHAPE_COUNT)

    def update_position(self, transform):
        omega_pos = HWR.beamline.diffractometer.omega.get_value() % 360
        _d = abs((self.get_centred_position().omega % 360) - omega_pos)

        if self.user_state == "HIDDEN":
            self.state = "HIDDEN"
            return

        if min(_d, 360 - _d) > self.shapes_hw_object.hide_grid_threshold:
            self.state = "HIDDEN"
        else:
            super().update_position(transform)
            self.state = "SAVED"

    def get_centred_position(self):
        return self.cp_list[0]

    def get_grid_range(self):
        return (
            float(self.cell_width * (self.num_cols - 1)),
            float(self.cell_height * (self.num_rows - 1)),
        )

    def get_num_lines(self):
        if self.cell_count_fun == "zig-zag":
            return self.num_rows
        if self.cell_count_fun == "inverse-zig-zag":
            return self.num_cols
        return self.num_rows

    def set_id(self, id_num):
        Shape.set_id(self, id_num)
        self.cp_list[0].index = self.name

    def set_result(self, result_data):
        self.result = result_data

    def get_result(self):
        return self.result

    def as_dict(self) -> dict:
        """Convert a shape to a dictionary."""
        d = Shape.as_dict(self)
        # replace cpos_list with the motor positions
        d["motor_positions"] = self.cp_list[0].as_dict()

        pixels_per_mm = HWR.beamline.diffractometer.get_pixels_per_mm()
        beam_pos = HWR.beamline.beam.get_beam_position_on_screen()
        size_x, size_y, shape, _label = HWR.beamline.beam.get_value()

        d["x1"] = -float((beam_pos[0] - d["screen_coord"][0]) / pixels_per_mm[0])
        d["y1"] = -float((beam_pos[1] - d["screen_coord"][1]) / pixels_per_mm[1])
        d["steps_x"] = d["num_cols"]
        d["steps_y"] = d["num_rows"]
        d["dx_mm"] = d["width"] / pixels_per_mm[0]
        d["dy_mm"] = d["height"] / pixels_per_mm[1]
        d["beam_width"] = size_x
        d["beam_height"] = size_y
        d["angle"] = 0

        return d



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.
"""
Microdiff with Exporter implementation of AbstartNState
Example xml file:
<object class="ExporterNState">
  <username>Fluorescence Detector</username>
  <exporter_address>wid30bmd2s:9001</exporter_address>
  <value_channel_name>FluoDetectorIsBack</value_channel_name>
  <state_channel_name>State</state_channel_name>
  <values>{"IN": False, "OUT": True}</values>
</object>
"""

from enum import Enum

from gevent import sleep

from mxcubecore.Command.exporter.ExporterStates import ExporterStates
from mxcubecore.HardwareObjects.abstract.AbstractNState import AbstractNState

__copyright__ = """ Copyright © 2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class ExporterNStateMockup(AbstractNState):
    """Microdiff with Exporter implementation of AbstartNState"""

    SPECIFIC_STATES = ExporterStates

    def __init__(self, name):
        AbstractNState.__init__(self, name)

    def init(self):
        """Initialise the device"""
        AbstractNState.init(self)
        value = self.VALUES.OUT
        self.update_state(self.STATES.READY)
        self.update_value(value)

    def _wait_ready(self, timeout=None):
        """Wait timeout seconds till status is ready.
        Args:
            timeout(float): Timeout [s]. None means infinite timeout.
        """
        sleep(0.5)

    def _update_state(self, state=None):
        """To be used to update the state when emitting the "update" signal.
        Args:
            state (str): optional state value
        Returns:
            (enum 'HardwareObjectState'): state.
        """
        if not state:
            state = self.get_state()
        else:
            state = self._value2state(state)

            return self.update_state(state)

    def _value2state(self, state):
        """Convert string state to HardwareObjectState enum value
        Args:
            state (str): the state
        Returns:
            (enum 'HardwareObjectState'): state
        """
        try:
            return self.SPECIFIC_STATES.__members__[state.upper()].value
        except (AttributeError, KeyError):
            return self.STATES.UNKNOWN

    def abort(self):
        """Stop the action."""
        pass

    def _set_value(self, value):
        """Set device to value
        Args:
            value (str, int, float or enum): Value to be set.
        """
        self.update_state(self.STATES.BUSY)

        sleep(0.5)

        if isinstance(value, Enum):
            if isinstance(value.value, (tuple, list)):
                value = value.value[0]
            else:
                value = value.value

        self.update_value(self.value_to_enum(value))
        self.update_state(self.STATES.READY)

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): Enum member, corresponding to the value or UNKNOWN.
        """
        return self._nominal_value



from mxcubecore.BaseHardwareObjects import HardwareObject


class BlueskyHttpServer(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

    def init(self):
        super().init()
        self.api = self.get_command_object("bluesky_http_server")
        self.api.username = "mnc-data"

    def execute_plan(self, plan_name, kwargs=None):
        response = self.api.execute_plan(plan_name=plan_name, kwargs=kwargs)
        response_content = response.json()
        if response_content["success"]:
            try:
                self.api.monitor_manager_state("executing_queue")
                self.api.monitor_manager_state("idle")
            except TimeoutError:
                self.log.exception("The Bluesky plan has timed out!")
        else:
            self.log.exception(response_content["msg"])



import atexit
import time

from pymba import Vimba

try:
    import cv2
except ImportError:
    pass

from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice
from mxcubecore.utils.qt_import import (
    QImage,
    QPixmap,
)


class VimbaVideo(AbstractVideoDevice):
    def __init__(self, name):
        AbstractVideoDevice.__init__(self, name)

        self.camera = None
        self.camera_id = None
        self.qimage = None
        self.qpixmap = None
        self.use_qt = False
        self.raw_image_dimensions = [1360, 1024]

    def init(self):
        # start Vimba
        self.camera_id = self.get_property("camera_id")
        self.use_qt = self.get_property("use_qt", True)

        atexit.register(self.close_camera)
        AbstractVideoDevice.init(self)

    def get_raw_image_size(self):
        return self.raw_image_dimensions

    def do_image_polling(self, sleep_time):
        with Vimba() as vimba:
            self.camera = vimba.camera(self.camera_id)
            self.camera.open(camera_access_mode=2, adjust_packet_size=False)
            self.frame = self.camera.new_frame()
            self.frame.announce()
            self.camera.start_capture()

            self.raw_image_dimensions = (
                self.camera.feature("Width").value,
                self.camera.feature("Height").value,
            )

            while True:
                self.frame.wait_for_capture(1000)
                self.frame.queue_for_capture()

                if self.use_qt:
                    self.qimage = QImage(
                        self.frame.buffer_data(),
                        self.raw_image_dimensions[0],
                        self.raw_image_dimensions[1],
                        QImage.Format_RGB888,
                    )
                    if self.cam_mirror is not None:
                        self.qimage = self.qimage.mirrored(
                            self.cam_mirror[0], self.cam_mirror[1]
                        )
                    if self.qpixmap is None:
                        self.qpixmap = QPixmap(self.qimage)
                    else:
                        self.qpixmap.convertFromImage(self.qimage)
                    self.emit("imageReceived", self.qpixmap)
                else:
                    image_data = cv2.cvtColor(
                        self.frame.buffer_data_numpy(), cv2.COLOR_BGR2RGBA
                    )
                    ret, im = cv2.imencode(".jpg", image_data)
                    self.emit(
                        "imageReceived",
                        im.tostring(),
                        self.frame.width,
                        self.frame.height,
                    )

                time.sleep(sleep_time)

    def get_new_image(self):
        return self.qimage

    def get_video_live(self):
        return True

    def close_camera(self):
        with Vimba() as vimba:
            self.camera.flush_capture_queue()
            self.camera.end_capture()
            self.camera.revoke_all_frames()
            vimba.shutdown()

    def start_camera(self):
        pass

    def change_owner(self):
        pass



"""LimaPilatus4Detector Class
Lima Tango Device Server implementation of the Dectris Pilatus4 Detector.
Differs from Eiger only in the energy threshold handling.
"""

import math

import requests

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.LimaEigerDetector import LimaEigerDetector


class LimaPilatus4Detector(LimaEigerDetector):
    def init(self):
        """Initialise the photon_energy and the thresholds"""
        super().init()

        eiger_device = self.get_property("eiger_device")

        for channel_name in ("threshold_energy", "threshold_energy2"):
            self.add_channel(
                {"type": "tango", "name": channel_name, "tangoname": eiger_device},
                channel_name,
            )
        # the threshold must be reset after setting the photon_energy
        self.set_energy_threshold(HWR.beamline.energy.get_value())

    def set_energy_threshold(self, energy):
        min_e = self.get_property("minE")
        energy = max(energy, min_e)

        working_energy_chan = self.get_channel_object("photon_energy")
        working_energy = working_energy_chan.get_value() / 1000.0
        if math.fabs(working_energy - energy) > 0.1:
            egy = int(energy * 1000.0)
            working_energy_chan.set_value(egy)

        threshold = self.get_channel_object("threshold_energy").get_value()
        # set the other thresholds to the same value as 1
        for chn in range(2, 5, 1):
            api_url = self.get_property("detector_api_url")
            if not api_url:
                raise RuntimeError("Cannot set detector energy threshold")  # noqa: EM101
            url = f"{api_url}/config/threshold/{chn}/energy"
            try:
                requests.put(url, '{"value": %f}' % threshold, timeout=120)
            except Exception:  # noqa: BLE001
                requests.put(url, '{"value": %f}' % threshold, timeout=120)



#
#  Project name: MXCuBE
#  https://github.com/mxcube.
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time
from math import sqrt

import gevent

try:
    import lucid3 as lucid
except ImportError:
    try:
        import lucid
    except ImportError:
        logging.warning(
            "Could not find autocentring library, automatic centring is disabled"
        )

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.GenericDiffractometer import GenericDiffractometer

__credits__ = ["EMBL Hamburg"]
__category__ = "General"


class EMBLMiniDiff(GenericDiffractometer):
    AUTOMATIC_CENTRING_IMAGES = 6
    CENTRING_METHOD_IMAGING = "3-click imaging"
    CENTRING_METHOD_IMAGING_N = "n-click imaging"

    def __init__(self, *args):
        GenericDiffractometer.__init__(self, *args)

        # Hardware objects ----------------------------------------------------
        self.omega_reference_motor = None
        self.centring_hwobj = None
        self.minikappa_correction_hwobj = None
        self.zoom_motor_hwobj = None

        # Channels and commands -----------------------------------------------
        self.chan_calib_x = None
        self.chan_calib_y = None
        self.chan_current_phase = None
        self.chan_head_type = None
        self.chan_fast_shutter_is_open = None
        self.chan_state = None
        self.chan_sync_move_motors = None
        self.chan_scintillator_position = None
        self.chan_capillary_position = None
        self.cmd_start_set_phase = None
        self.cmd_start_auto_focus = None
        self.cmd_get_omega_scan_limits = None
        self.cmd_save_centring_positions = None

        # Internal values -----------------------------------------------------
        self.use_sc = False
        self.omega_reference_par = None
        self.omega_reference_pos = [0, 0]
        self.imaging_pixels_per_mm = [0, 0]

        self.current_phase = None

    def init(self):
        GenericDiffractometer.init(self)
        self.centring_status = {"valid": False}

        self.chan_state = self.get_channel_object("State")
        self.current_state = self.chan_state.get_value()
        self.chan_state.connect_signal("update", self.state_changed)

        self.chan_status = self.get_channel_object("Status")
        self.chan_status.connect_signal("update", self.status_changed)

        self.chan_calib_x = self.get_channel_object("CoaxCamScaleX")
        self.chan_calib_y = self.get_channel_object("CoaxCamScaleY")
        self.update_pixels_per_mm()

        self.chan_head_type = self.get_channel_object("HeadType")
        self.head_type = self.chan_head_type.get_value()

        self.chan_current_phase = self.get_channel_object("CurrentPhase")
        self.connect(self.chan_current_phase, "update", self.current_phase_changed)

        self.chan_fast_shutter_is_open = self.get_channel_object("FastShutterIsOpen")
        self.chan_fast_shutter_is_open.connect_signal(
            "update", self.fast_shutter_state_changed
        )

        self.chan_scintillator_position = self.get_channel_object(
            "ScintillatorPosition"
        )
        self.chan_capillary_position = self.get_channel_object("CapillaryPosition")

        self.cmd_start_set_phase = self.get_command_object("startSetPhase")
        self.cmd_start_auto_focus = self.get_command_object("startAutoFocus")
        self.cmd_get_omega_scan_limits = self.get_command_object(
            "getOmegaMotorDynamicScanLimits"
        )
        self.cmd_save_centring_positions = self.get_command_object(
            "saveCentringPositions"
        )

        self.centring_hwobj = self.get_object_by_role("centring")
        self.imaging_centring_hwobj = self.get_object_by_role("imaging-centring")
        self.minikappa_correction_hwobj = self.get_object_by_role(
            "minikappa_correction"
        )
        self.zoom_motor_hwobj = self.get_object_by_role("zoom")
        self.connect(self.zoom_motor_hwobj, "valueChanged", self.zoom_position_changed)
        self.connect(
            self.zoom_motor_hwobj,
            "predefinedPositionChanged",
            self.zoom_motor_predefined_position_changed,
        )
        self.connect(self.motor_hwobj_dict["phi"], "valueChanged", self.phi_motor_moved)
        self.connect(
            self.motor_hwobj_dict["phiy"], "valueChanged", self.phiy_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["phiz"], "valueChanged", self.phiz_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa"], "valueChanged", self.kappa_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["kappa_phi"],
            "valueChanged",
            self.kappa_phi_motor_moved,
        )
        self.connect(
            self.motor_hwobj_dict["sampx"], "valueChanged", self.sampx_motor_moved
        )
        self.connect(
            self.motor_hwobj_dict["sampy"], "valueChanged", self.sampy_motor_moved
        )

        self.kappa_motor_moved(self.motor_hwobj_dict["kappa"].get_value())
        self.kappa_phi_motor_moved(self.motor_hwobj_dict["kappa_phi"].get_value())

        self.omega_reference_par = eval(self.get_property("omega_reference"))
        self.omega_reference_motor = self.get_object_by_role(
            self.omega_reference_par["motor_name"]
        )
        self.connect(
            self.omega_reference_motor,
            "valueChanged",
            self.omega_reference_motor_moved,
        )

        # self.use_sc = self.get_property("use_sample_changer")
        self.imaging_pixels_per_mm = [1, 1]
        self.centring_methods[EMBLMiniDiff.CENTRING_METHOD_IMAGING] = (
            self.start_imaging_centring
        )
        self.centring_methods[EMBLMiniDiff.CENTRING_METHOD_IMAGING_N] = (
            self.start_imaging_centring_n
        )

    def use_sample_changer(self):
        """Returns true if sample changer is used

        :return: bool
        """
        return not self.in_plate_mode

    def beam_position_changed(self, value):
        self.beam_position = value

    def state_changed(self, state):
        # self.log.debug("State changed: %s" % str(state))
        self.current_state = state
        self.emit("minidiffStateChanged", (self.current_state))
        self.emit("minidiffStatusChanged", (self.current_state))

    def status_changed(self, state):
        self.emit("statusMessage", ("diffractometer", state, "busy"))

    def zoom_position_changed(self, value):
        self.update_pixels_per_mm()
        self.current_motor_positions["zoom"] = value
        self.refresh_omega_reference_position()

    def zoom_motor_predefined_position_changed(self, position_name, offset):
        self.update_pixels_per_mm()
        self.emit("zoomMotorPredefinedPositionChanged", (position_name, offset))

    def omega_reference_add_constraint(self):
        """
        Descript. :
        """
        if self.omega_reference_par is None or self.beam_position is None:
            return
        if self.omega_reference_par["camera_axis"].lower() == "x":
            on_beam = (
                (self.beam_position[0] - self.zoom_centre["x"])
                * self.omega_reference_par["direction"]
                / self.pixels_per_mm_x
                + self.omega_reference_par["position"]
            )  # fmt: skip
        else:
            on_beam = (
                (self.beam_position[1] - self.zoom_centre["y"])
                * self.omega_reference_par["direction"]
                / self.pixels_per_mm_y
                + self.omega_reference_par["position"]
            )  # fmt: skip
        self.centring_hwobj.appendMotorConstraint(self.omega_reference_motor, on_beam)

    def omega_reference_motor_moved(self, pos):
        if self.omega_reference_par["camera_axis"].lower() == "x":
            pos = (
                self.omega_reference_par["direction"]
                * (pos - self.omega_reference_par["position"])
                * self.pixels_per_mm_x
                + self.zoom_centre["x"]
            )
            self.reference_pos = (pos, -10)
        else:
            pos = (
                self.omega_reference_par["direction"]
                * (pos - self.omega_reference_par["position"])
                * self.pixels_per_mm_y
                + self.zoom_centre["y"]
            )
            self.reference_pos = (-10, pos)
        self.emit("omegaReferenceChanged", (self.reference_pos,))

    def fast_shutter_state_changed(self, is_open):
        self.fast_shutter_is_open = is_open
        if is_open:
            msg = "Opened"
        else:
            msg = "Closed"
        self.emit("minidiffShutterStateChanged", (self.fast_shutter_is_open, msg))

    def phi_motor_moved(self, pos):
        self.current_motor_positions["phi"] = pos
        self.emit("phiMotorMoved", pos)

    def phiy_motor_moved(self, pos):
        self.current_motor_positions["phiy"] = pos

    def phiz_motor_moved(self, pos):
        self.current_motor_positions["phiz"] = pos

    def sampx_motor_moved(self, pos):
        self.current_motor_positions["sampx"] = pos

    def sampy_motor_moved(self, pos):
        self.current_motor_positions["sampy"] = pos

    def kappa_motor_moved(self, pos):
        self.current_motor_positions["kappa"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaMotorMoved", pos)

    def kappa_phi_motor_moved(self, pos):
        self.current_motor_positions["kappa_phi"] = pos
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()
        self.emit("kappaPhiMotorMoved", pos)

    def refresh_omega_reference_position(self):
        if self.omega_reference_motor is not None:
            reference_pos = self.omega_reference_motor.get_value()
            self.omega_reference_motor_moved(reference_pos)

    def update_pixels_per_mm(self, *args):
        self.pixels_per_mm_x = 1.0 / self.chan_calib_x.get_value()
        self.pixels_per_mm_y = 1.0 / self.chan_calib_y.get_value()
        self.emit("pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y),))

    def set_phase(self, phase, timeout=80):
        """Sets diffractometer to the selected phase.
        In the plate mode before going to or away from
        Transfer or Beam location phase if needed then detector
        is moved to the safe distance to avoid collision.
        """
        if phase != self.get_current_phase():
            logging.getLogger("GUI").warning(
                "Diffractometer: Setting %s phase. Please wait..." % phase
            )

        if self.in_plate_mode and (
            phase
            in (GenericDiffractometer.PHASE_TRANSFER, GenericDiffractometer.PHASE_BEAM)
            or self.current_phase
            in (GenericDiffractometer.PHASE_TRANSFER, GenericDiffractometer.PHASE_BEAM)
        ):
            detector_distance = HWR.beamline.detector.distance.get_value()
            self.log.debug(
                "Diffractometer current phase: %s " % self.current_phase
                + "selected phase: %s" % phase
                + "detector distance: %d mm" % detector_distance
            )
            if detector_distance < 350:
                logging.getLogger("GUI").info("Moving detector to safe distance")
                HWR.beamline.detector.distance.set_value(350, timeout=20)

        if timeout is not None:
            _start = time.time()
            self.cmd_start_set_phase(phase)
            gevent.sleep(5)
            with gevent.Timeout(
                timeout, Exception("Timeout waiting for phase %s" % phase)
            ):
                while phase != self.chan_current_phase.get_value():
                    gevent.sleep(0.1)
            self.wait_device_ready(30)
            self.wait_device_ready(30)
            _howlong = time.time() - _start
            if _howlong > 11.0:
                self.log.error(
                    "Changing phase to %s took %.1f seconds" % (phase, _howlong)
                )
        else:
            self.cmd_start_set_phase(phase)

    def start_auto_focus(self, timeout=None):
        if timeout:
            self.ready_event.clear()
            set_phase_task = gevent.spawn(
                self.execute_server_task, self.cmd_start_auto_focus(), timeout
            )
            self.ready_event.wait()
            self.ready_event.clear()
        else:
            self.cmd_start_auto_focus()

    def emit_diffractometer_moved(self, *args):
        self.emit("diffractometerMoved", ())

    def invalidate_centring(self):
        if self.current_centring_procedure is None and self.centring_status["valid"]:
            self.centring_status = {"valid": False}
            self.emit_progress_message("")
            self.emit("centringInvalid", ())

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        self.centring_hwobj.initCentringProcedure()
        self.centring_hwobj.appendCentringDataPoint(
            {
                "X": (x - self.beam_position[0]) / self.pixels_per_mm_x,
                "Y": (y - self.beam_position[1]) / self.pixels_per_mm_y,
            }
        )
        self.omega_reference_add_constraint()
        pos = self.centring_hwobj.centeredPosition()
        if return_by_names:
            pos = self.convert_from_obj_to_name(pos)
        return pos

    def move_to_beam(self, x, y, omega=None):
        """Creates a new centring point based on all motors positions"""
        if self.current_phase != "BeamLocation":
            GenericDiffractometer.move_to_beam(self, x, y, omega)
        else:
            self.log.debug(
                "Diffractometer: Move to screen"
                + " position disabled in BeamLocation phase."
            )

    def manual_centring(self):
        self.centring_hwobj.initCentringProcedure()
        for click in range(3):
            self.user_clicked_event = gevent.event.AsyncResult()
            x, y = self.user_clicked_event.get()
            self.centring_hwobj.appendCentringDataPoint(
                {
                    "X": (x - self.beam_position[0]) / self.pixels_per_mm_x,
                    "Y": (y - self.beam_position[1]) / self.pixels_per_mm_y,
                }
            )
            if self.in_plate_mode:
                dynamic_limits = self.get_osc_limits()
                if click == 0:
                    self.motor_hwobj_dict["phi"].set_value(dynamic_limits[0] + 0.5)
                elif click == 1:
                    self.motor_hwobj_dict["phi"].set_value(dynamic_limits[1] - 0.5)
                elif click == 2:
                    self.motor_hwobj_dict["phi"].set_value(
                        (dynamic_limits[0] + dynamic_limits[1]) / 2.0
                    )
            else:
                if click < 2:
                    self.motor_hwobj_dict["phi"].set_value_relative(90)
        self.omega_reference_add_constraint()
        return self.centring_hwobj.centeredPosition(return_by_name=False)

    def automatic_centring(self):
        """Automatic centring procedure. Rotates n times and executes
        centring algorithm. Optimal scan position is detected.
        """
        self.wait_device_ready(20)
        surface_score_list = []
        self.zoom_motor_hwobj.move_to_position("Zoom 1")
        self.centring_hwobj.initCentringProcedure()
        for image in range(EMBLMiniDiff.AUTOMATIC_CENTRING_IMAGES):
            x, y, score = self.find_loop()
            if x > 0 and y > 0:
                self.centring_hwobj.appendCentringDataPoint(
                    {
                        "X": (x - self.beam_position[0]) / self.pixels_per_mm_x,
                        "Y": (y - self.beam_position[1]) / self.pixels_per_mm_y,
                    }
                )
            surface_score_list.append(score)
            self.motor_hwobj_dict["phi"].set_value_relative(
                360.0 / EMBLMiniDiff.AUTOMATIC_CENTRING_IMAGES
            )
            gevent.sleep(0.01)
            self.wait_device_ready(15)
        self.omega_reference_add_constraint()
        centred_pos_dir = self.centring_hwobj.centeredPosition(return_by_name=False)
        # self.emit("newAutomaticCentringPoint", centred_pos_dir)

        return centred_pos_dir

    def start_imaging_centring(self, sample_info=None, wait_result=None):
        self.emit_progress_message("Imaging based 3 click centring...")
        self.current_centring_procedure = gevent.spawn(self.imaging_centring)
        self.current_centring_procedure.link(self.centring_done)

    def start_imaging_centring_n(self, sample_info=None, wait_result=None):
        self.emit_progress_message("Imaging based n click centring...")
        self.current_centring_procedure = gevent.spawn(self.imaging_centring_n)
        self.current_centring_procedure.link(self.centring_done)

    def imaging_centring(self):
        self.imaging_centring_hwobj.initCentringProcedure()
        for click in range(3):
            self.user_clicked_event = gevent.event.AsyncResult()
            x, y = self.user_clicked_event.get()
            self.imaging_centring_hwobj.appendCentringDataPoint(
                {
                    "X": (x - 1024.0) / self.imaging_pixels_per_mm[0],
                    "Y": (y - 1024.0) / self.imaging_pixels_per_mm[1],
                }
            )
            if click < 2:
                self.motor_hwobj_dict["phi"].set_value_relative(90)
            # print "rotate omega"
        # self.omega_reference_add_constraint()
        return self.imaging_centring_hwobj.centeredPosition(return_by_name=False)

    def imaging_centring_n(self):
        self.imaging_centring_hwobj.initCentringProcedure()
        while True:
            self.user_clicked_event = gevent.event.AsyncResult()
            x, y = self.user_clicked_event.get()
            self.imaging_centring_hwobj.appendCentringDataPoint(
                {
                    "X": (x - 1024.0) / self.imaging_pixels_per_mm[0],
                    "Y": (y - 1024.0) / self.imaging_pixels_per_mm[1],
                }
            )
        # self.omega_reference_add_constraint()
        return self.imaging_centring_hwobj.centeredPosition(return_by_name=False)

    def motor_positions_to_screen(self, centred_positions_dict):
        c = centred_positions_dict

        # kappa = self.current_motor_positions["kappa"]
        # phi = self.current_motor_positions["kappa_phi"]

        kappa = self.motor_hwobj_dict["kappa"].get_value()
        phi = self.motor_hwobj_dict["kappa_phi"].get_value()
        # IK TODO remove this director call

        if (c["kappa"], c["kappa_phi"]) != (
            kappa,
            phi,
        ) and self.minikappa_correction_hwobj is not None:
            # c['sampx'], c['sampy'], c['phiy']
            c["sampx"], c["sampy"], c["phiy"] = self.minikappa_correction_hwobj.shift(
                c["kappa"],
                c["kappa_phi"],
                [c["sampx"], c["sampy"], c["phiy"]],
                kappa,
                phi,
            )
        xy = self.centring_hwobj.centringToScreen(c)
        if xy:
            x = (xy["X"] + c["beam_x"]) * self.pixels_per_mm_x + self.zoom_centre["x"]
            y = (xy["Y"] + c["beam_y"]) * self.pixels_per_mm_y + self.zoom_centre["y"]
            return x, y

    def move_to_centred_position(self, centred_position):
        if self.current_phase != "BeamLocation":
            try:
                x, y = centred_position.beam_x, centred_position.beam_y
                dx = (
                    self.beam_position[0] - self.zoom_centre["x"]
                ) / self.pixels_per_mm_x - x
                dy = (
                    self.beam_position[1] - self.zoom_centre["y"]
                ) / self.pixels_per_mm_y - y
                motor_pos = {
                    self.motor_hwobj_dict["sampx"]: centred_position.sampx,
                    self.motor_hwobj_dict["sampy"]: centred_position.sampy,
                    self.motor_hwobj_dict["phi"]: centred_position.phi,
                    self.motor_hwobj_dict["phiy"]: (
                        centred_position.phiy
                        + self.centring_hwobj.camera2alignmentMotor(
                            self.motor_hwobj_dict["phiy"], {"X": dx, "Y": dy}
                        )
                    ),
                    self.motor_hwobj_dict["phiz"]: (
                        centred_position.phiz
                        + self.centring_hwobj.camera2alignmentMotor(
                            self.motor_hwobj_dict["phiz"], {"X": dx, "Y": dy}
                        )
                    ),
                    self.motor_hwobj_dict["kappa"]: centred_position.kappa,
                    self.motor_hwobj_dict["kappa_phi"]: centred_position.kappa_phi,
                }
                self.move_to_motors_positions(motor_pos)
            except BaseException:
                logging.exception("Could not move to centred position")
        else:
            self.log.debug("Move to centred position disabled in BeamLocation phase.")

    def move_kappa_and_phi(self, kappa=None, kappa_phi=None, wait=False):
        return gevent.spawn(self.move_kappa_and_phi_procedure, kappa, kappa_phi)

    # @task
    def move_kappa_and_phi_procedure(self, new_kappa=None, new_kappa_phi=None):
        kappa = self.motor_hwobj_dict["kappa"].get_value()
        kappa_phi = self.motor_hwobj_dict["kappa_phi"].get_value()

        if new_kappa is None:
            new_kappa = kappa
        if new_kappa_phi is None:
            new_kappa_phi = kappa_phi

        motor_pos_dict = {}

        if (kappa, kappa_phi) != (
            new_kappa,
            new_kappa_phi,
        ) and self.minikappa_correction_hwobj is not None:
            sampx = self.motor_hwobj_dict["sampx"].get_value()
            sampy = self.motor_hwobj_dict["sampy"].get_value()
            phiy = self.motor_hwobj_dict["phiy"].get_value()
            new_sampx, new_sampy, new_phiy = self.minikappa_correction_hwobj.shift(
                kappa, kappa_phi, [sampx, sampy, phiy], new_kappa, new_kappa_phi
            )

            motor_pos_dict[self.motor_hwobj_dict["kappa"]] = new_kappa
            motor_pos_dict[self.motor_hwobj_dict["kappa_phi"]] = new_kappa_phi
            motor_pos_dict[self.motor_hwobj_dict["sampx"]] = new_sampx
            motor_pos_dict[self.motor_hwobj_dict["sampy"]] = new_sampy
            motor_pos_dict[self.motor_hwobj_dict["phiy"]] = new_phiy

            self.move_motors(motor_pos_dict, timeout=30)

    def convert_from_obj_to_name(self, motor_pos):
        motors = {}
        for motor_role in (
            "phiy",
            "phiz",
            "sampx",
            "sampy",
            "zoom",
            "phi",
            "focus",
            "kappa",
            "kappa_phi",
        ):
            mot_obj = self.get_object_by_role(motor_role)
            try:
                motors[motor_role] = motor_pos[mot_obj]
            except KeyError:
                motors[motor_role] = mot_obj.get_value()
        motors["beam_x"] = (
            self.beam_position[0] - self.zoom_centre["x"]
        ) / self.pixels_per_mm_y
        motors["beam_y"] = (
            self.beam_position[1] - self.zoom_centre["y"]
        ) / self.pixels_per_mm_x
        return motors

    def visual_align(self, point_1, point_2):
        if self.in_plate_mode:
            self.log.info("EMBLMiniDiff: Visual align not available in Plate mode")
        else:
            t1 = [point_1.sampx, point_1.sampy, point_1.phiy]
            t2 = [point_2.sampx, point_2.sampy, point_2.phiy]
            kappa = self.motor_hwobj_dict["kappa"].get_value()
            phi = self.motor_hwobj_dict["kappa_phi"].get_value()
            (
                new_kappa,
                new_phi,
                (
                    new_sampx,
                    new_sampy,
                    new_phiy,
                ),
            ) = self.minikappa_correction_hwobj.alignVector(t1, t2, kappa, phi)
            self.move_to_motors_positions(
                {
                    self.motor_hwobj_dict["kappa"]: new_kappa,
                    self.motor_hwobj_dict["kappa_phi"]: new_phi,
                    self.motor_hwobj_dict["sampx"]: new_sampx,
                    self.motor_hwobj_dict["sampy"]: new_sampy,
                    self.motor_hwobj_dict["phiy"]: new_phiy,
                }
            )

    def force_emit_signals(self):
        GenericDiffractometer.force_emit_signals(self)
        self.emit("kappaMotorMoved", (self.current_motor_positions["kappa"],))
        self.emit("kappaPhiMotorMoved", (self.current_motor_positions["kappa_phi"],))
        self.emit("minidiffPhaseChanged", (self.current_phase,))
        self.emit("omegaReferenceChanged", (self.reference_pos,))
        self.emit("minidiffShutterStateChanged", (self.fast_shutter_is_open,))

    def toggle_fast_shutter(self):
        if self.chan_fast_shutter_is_open is not None:
            self.chan_fast_shutter_is_open.set_value(not self.fast_shutter_is_open)

    def find_loop(self):
        image_array = self.camera_hwobj.get_snapshot(return_as_array=True)
        (info, x, y) = lucid.find_loop(image_array)
        surface_score = 10
        return x, y, surface_score

    def move_omega(self, angle):
        self.motor_hwobj_dict["phi"].set_value(angle, timeout=5)

    def move_omega_relative(self, relative_angle, timeout=5):
        self.motor_hwobj_dict["phi"].set_value_relative(relative_angle, timeout=timeout)

    def close_kappa(self):
        gevent.spawn(self.close_kappa_task)

    def close_kappa_task(self):
        """Close kappa task"""
        self.log.debug("Diffractometer: Closing Kappa started...")
        self.move_kappa_and_phi_procedure(0, 270)
        self.wait_device_ready(180)
        self.motor_hwobj_dict["kappa"].home()
        self.wait_device_ready(60)
        self.log.debug("Diffractometer: Done Closing Kappa")

    def set_zoom(self, position):
        self.zoom_motor_hwobj.move_to_position(position)

    def get_point_from_line(self, point_one, point_two, frame_num, frame_total):
        """
        method used to get a new motor position based on a position
        between two positions. As arguments both motor positions are
        given. frame_num and frame_total is used estimate new point position
        Helical line goes from point_one to point_two.
        In this direction also new position is estimated
        """
        new_point = {}
        point_one = point_one.as_dict()
        point_two = point_two.as_dict()
        for motor in point_one.keys():
            new_motor_pos = point_one[motor] + (
                point_two[motor] - point_one[motor]
            ) * frame_num / float(frame_total)
            new_point[motor] = new_motor_pos
        return new_point

    def get_osc_limits(self, speed=None):
        if speed:
            limits = self.cmd_get_omega_scan_limits(speed)
        else:
            limits = self.motor_hwobj_dict["phi"].get_dynamic_limits()
        return (min(limits), max(limits))

    def get_osc_max_speed(self):
        return self.motor_hwobj_dict["phi"].get_max_speed()

    def get_scan_limits(self, num_images=0, exp_time=0.001343):
        """
        Gets scan limits. Necessary for example in the plate mode
        where osc range is limited
        """

        """
        if num_images==0:
            try:
                return (155.7182, 204.8366)
                limits = self.cmd_get_omega_scan_limits(0)
                return (min(limits) + 0.01, max(limits) - 0.01)
            except:
                return (None, None)
        """

        total_exposure_time = num_images * exp_time
        a = 0.002
        b = 0.1537
        w0 = 155.7182
        w1 = 204.8366  # was 196 not to shadow laser

        if num_images == 0:
            return (w0, w1)

        speed = (
            (
                -2 * b
                - total_exposure_time
                + sqrt((2 * b + total_exposure_time) ** 2 - 8 * a * (w0 - w1))
            )
            / 4
            / a
        )
        if speed < 0:
            return None, None
        else:
            delta = a * speed**2 + b * speed

            return (w0 + delta, w1 - delta)

        """
        if speed is not None:
            try:
                limits = self.cmd_get_omega_scan_limits(speed)
                return (min(limits), max(limits)), None
            except:
                return None, None

        total_exposure_time = num_images * exp_time
        tmp = self.cmd_get_omega_scan_limits(0)
        max_speed = self.get_osc_max_speed()
        w0 = tmp[1]
        w1 = tmp[0]
        x1 = 10
        x2 = 50

        c1 = min(self.cmd_get_omega_scan_limits(x1)) - w0
        c2 = min(self.cmd_get_omega_scan_limits(x2)) - w0

        a = -(c2 * x1 - c1 * x2)/(x1 * x2 * (x1 -x2))
        b = -(-c2 * pow(x1, 2) + c1 * pow(x2, 2))/(x1 *x2 * (x1 - x2))

        result_speed = (-2*b-total_exposure_time+sqrt((2*b+total_exposure_time)**2-8*a*(w0-w1))) /4/a
        if result_speed < 0:
            return (None, None), None
        elif result_speed > max_speed:
            delta = a * max_speed**2 + b * max_speed
            #total_exposure_time = total_exposure_time * result_speed / max_speed
            total_exposure_time = (w1-w0-2*delta) / (max_speed - 0.1)
        else:
            delta = a * result_speed**2 + b * result_speed

        print "get scan limits 2 ", (w0 + delta, w1 - delta)

        return (w0 + delta, w1 - delta), total_exposure_time / num_images
        """

    def get_scintillator_position(self):
        return self.chan_scintillator_position.get_value()

    def set_scintillator_position(self, position):
        self.chan_scintillator_position.set_value(position)
        with gevent.Timeout(5, Exception("Timeout waiting for scintillator position")):
            while position != self.get_scintillator_position():
                gevent.sleep(0.01)

    def get_capillary_position(self):
        return self.chan_capillary_position.get_value()

    def set_capillary_position(self, position):
        self.chan_capillary_position.set_value(position)
        with gevent.Timeout(60, Exception("Timeout waiting for capillary position")):
            while position != self.get_capillary_position():
                gevent.sleep(0.01)

    def zoom_in(self):
        self.zoom_motor_hwobj.zoom_in()

    def zoom_out(self):
        self.zoom_motor_hwobj.zoom_out()

    def save_centring_positions(self):
        self.cmd_save_centring_positions()
        # self.set_zoom("Zoom 7")

    def move_sample_out(self):
        self.motor_hwobj_dict["phiy"].set_value_relative(-2, wait=True, timeout=5)

    def move_sample_in(self):
        self.motor_hwobj_dict["phiy"].set_value_relative(2, wait=True, timeout=5)



import logging
import os
import shutil

from mxcubecore import HardwareRepository as HWR
from mxcubecore.TaskUtils import task

from .ESRFMultiCollect import ESRFMultiCollect


class MD2MultiCollect(ESRFMultiCollect):
    def __init__(self, name):
        ESRFMultiCollect.__init__(self, name)
        self.fast_characterisation = None

    @task
    def data_collection_hook(self, data_collect_parameters):
        ESRFMultiCollect.data_collection_hook(self, data_collect_parameters)
        self._detector.shutterless = data_collect_parameters["shutterless"]

        if hasattr(HWR.beamline.sample_changer, "get_crystal_id"):
            comment = HWR.beamline.sample_changer.get_crystal_id()
            data_collect_parameters["comment"] = comment

    @task
    def get_beam_size(self):
        _width, _height, _, _ = HWR.beamline.beam.get_value()
        return _width, _height

    @task
    def get_slit_gaps(self):
        self.get_object_by_role("controller")

        return (None, None)

    @task
    def get_beam_shape(self):
        return HWR.beamline.beam.get_value()[2].name

    def get_resolution_at_corner(self):
        return HWR.beamline.resolution.get_value_at_corner()

    def ready(*motors):
        return not any([m.motorIsMoving() for m in motors])

    @task
    def move_motors(self, motors_to_move_dict):
        # We do not want to modify the input dict
        motor_positions_copy = motors_to_move_dict.copy()
        diffr = HWR.beamline.diffractometer
        for tag in ("kappa", "kappa_phi", "zoom"):
            if tag in motor_positions_copy:
                del motor_positions_copy[tag]

        diffr.set_value_motors(motor_positions_copy, simultaneous=True)

    @task
    def take_crystal_snapshots(self, number_of_snapshots, image_path_list=[]):
        HWR.beamline.sample_view.take_acq_snapshot(image_path_list)

    def do_prepare_oscillation(self, *args, **kwargs):
        diffr = HWR.beamline.diffractometer

        # send again the command as MD2 software only handles one
        # centered position!!
        # has to be where the motors are and before changing the phase
        # diffr.get_command_object("save_centring_positions")()

        # switch on the front light
        diffr.frontlightswitch.set_value(diffr.frontlightswitch.VALUES.IN)
        # diffr.frontlight.set_value(2)

        diffr.backlightswitch.disable()
        # move to DataCollection phase
        logging.getLogger("user_level_log").info("Moving MD2 to DataCollection")
        # AB next line to speed up the data collection
        diffr.set_phase(diffr.get_phase_enum.COLLECT, timeout=0)

    @task
    def data_collection_cleanup(self):
        HWR.beamline.diffractometer.wait_status_ready(10)
        self.close_fast_shutter()
        HWR.beamline.diffractometer.backlightswitch.enable()

    @task
    def oscil(self, start, end, exptime, number_of_images, wait=True):
        diffr = HWR.beamline.diffractometer
        # make sure the diffractometer is ready to do the scan
        diffr.wait_status_ready(100)
        if self.helical:
            diffr.do_line_scan(start, end, exptime, number_of_images, self.helical_pos)
        elif self.mesh:
            det = HWR.beamline.detector
            latency_time = det.get_property("latecy_time_mesh") or det.get_deadtime()
            sequence_trigger = self.get_property("lima_sequnce_trigger_mode") or False

            if sequence_trigger:
                msg = "Using LIMA sequnce trigger mode for Eiger"
                self.log.info(msg)
                mesh_total_nb_frames = self.mesh_num_lines
            else:
                mesh_total_nb_frames = self.mesh_total_nb_frames

            diffr.do_mesh_scan(
                start,
                end,
                exptime,
                latency_time,
                self.mesh_num_lines,
                mesh_total_nb_frames,
                self.mesh_center,
                self.mesh_range,
            )
        elif self.fast_characterisation:
            nb_frames = 10
            nb_scan = 4
            angle = 90
            exptime *= 10
            scan_range = (end - start) * 10
            diffr.do_characterisation_scan(
                start,
                scan_range,
                nb_frames,
                exptime,
                nb_scan,
                angle,
            )
        else:
            diffr.do_oscillation_scan(start, end, exptime, number_of_images)

    @task
    def prepare_acquisition(
        self, take_dark, start, osc_range, exptime, npass, number_of_images, comment=""
    ):
        if self.fast_characterisation:
            number_of_images *= 40
        ext_gate = self.mesh or self.fast_characterisation

        self._detector.prepare_acquisition(
            take_dark,
            start,
            osc_range,
            exptime,
            npass,
            number_of_images,
            comment,
            ext_gate,
            self.mesh_num_lines,
        )

    def open_fast_shutter(self):
        fs = HWR.beamline.fast_shutter
        fs.set_value(fs.VALUES.OPEN)

    def close_fast_shutter(self):
        fs = HWR.beamline.fast_shutter
        fs.set_value(fs.VALUES.CLOSED)

    def set_helical(self, helical_on):
        self.helical = helical_on

    def set_helical_pos(self, helical_oscil_pos):
        self.helical_pos = helical_oscil_pos

    # specifies the next scan will be a mesh scan
    def set_mesh(self, mesh_on):
        self.mesh = mesh_on

    def set_mesh_scan_parameters(
        self, num_lines, total_nb_frames, mesh_center_param, mesh_range_param
    ):
        """
        sets the mesh scan parameters :
         - vertcal range
         - horizontal range
         - nb lines
         - nb frames per line
         - invert direction (boolean)  # NOT YET DONE
        """
        self.mesh_num_lines = num_lines
        self.mesh_total_nb_frames = total_nb_frames
        self.mesh_range = mesh_range_param
        self.mesh_center = mesh_center_param

    def set_fast_characterisation(self, value=False):
        self.fast_characterisation = value

    def get_cryo_temperature(self):
        return 0

    @task
    def prepare_intensity_monitors(self):
        return

    def get_beam_centre(self):
        pixel_x, pixel_y = HWR.beamline.detector.get_pixel_size()
        bcx, bcy = HWR.beamline.detector.get_beam_position()
        return [bcx * pixel_x, bcy * pixel_y]

    @task
    def write_input_files(self, datacollection_id):
        # copy *geo_corr.cbf* files to process directory
        try:
            process_dir = os.path.join(self.xds_directory, "..")
            raw_process_dir = os.path.join(self.raw_data_input_file_dir, "..")
            for dir in (process_dir, raw_process_dir):
                for filename in ("x_geo_corr.cbf.bz2", "y_geo_corr.cbf.bz2"):
                    dest = os.path.join(dir, filename)
                    if os.path.exists(dest):
                        continue
                    shutil.copyfile(
                        os.path.join(
                            self.get_property("template_file_directory"), filename
                        ),
                        dest,
                    )
        except Exception:
            logging.exception("Exception happened while copying geo_corr files")

        return ESRFMultiCollect.write_input_files(self, datacollection_id)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
MicrodiffScintillator.

Example xml file:

.. code-block:: xml

  <object class="MicrodiffScintillator">
    <username>Scintilator</username>
    <exporter_address>xxx:9001</exporter_address>
    <value_channel_name>ScintillatorPosition</value_channel_name>
    <values>{"IN": "SCINTILLATOR", "OUT": "PARK"}</values>
    <state_channel_name>HardwareState</state_channel_name>
    <object href="/detector_cover" role="detector_cover"/>
  </object>
"""

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractDiffractometer import (
    DiffractometerPhase,
)
from mxcubecore.HardwareObjects.ExporterNState import ExporterNState

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class MicrodiffScintillator(ExporterNState):
    """Microdiff Scintillator class"""

    def _set_value(self, value):
        """Set device to value
        Args:
            value (str, int, float or enum): Value to be set.
        """
        # use phase when setting in
        if value == self.VALUES.IN:
            # use phase change instead of in
            HWR.beamline.diffractometer.set_phase(DiffractometerPhase.SEE_BEAM)
            # close the detecor cover
            detcover = self.get_object_by_role("detector_cover")
            if detcover:
                detcover.set_value(detcover.detector_cover.VALUES.CLOSE)
        else:
            super()._set_value(value)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
DoorInterlockMockup
"""

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["MXCuBE collaboration"]


class DoorInterlockMockup(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self.door_interlock_state = None

    def init(self):
        self.door_interlock_state = self.STATES.READY

    def connected(self):
        self.set_is_ready(True)

    def disconnected(self):
        self.set_is_ready(False)

    def door_is_interlocked(self):
        return self.door_interlock_state in [self.STATES.READY]

    def get_state(self):
        return self.door_interlock_state

    def unlock_door_interlock(self):
        if self.door_interlock_state == "locked_active":
            self.door_interlock_state = "unlocked"
        self.emit("doorInterlockStateChanged", self.door_interlock_state, "")
        self.emit("hutchTrigger", self.door_interlock_state == "unlocked")



"""
Utility functions for converting between different units.
"""

#
# time units
#


def us_to_sec(us: float) -> float:
    """
    convert microseconds (μs) to seconds
    """
    return us / 1_000_000.0


def ms_to_sec(ms: float) -> float:
    """
    convert milliseconds (ms) to seconds
    """
    return ms / 1000.0


def sec_to_us(sec: float) -> float:
    """
    convert seconds to microseconds (μs)
    """
    return sec * 1_000_000.0


def sec_to_hour(sec: float) -> float:
    """
    convert seconds to hours
    """
    return sec / (60 * 60)


#
# energy units
#


def ev_to_kev(ev: float) -> float:
    """
    convert eV value to KeV value
    """
    return ev / 1000.0


def kev_to_ev(ev: float) -> float:
    """
    convert KeV value to eV value
    """
    return ev * 1000.0


#
# length units
#


def meter_to_mm(meters: float) -> float:
    """
    convert meters to millimeters (mm)
    """
    return meters * 1000.0


def mm_to_meter(millimeters: float) -> float:
    """
    convert millimeters (mm) to meters
    """
    return millimeters / 1000.0


def um_to_mm(micrometers: float) -> float:
    """
    convert micrometers (μm) to millimeters
    """
    return micrometers / 1000.0


#
# current units
#


def A_to_mA(amp: float) -> float:
    """
    convert Ampere (A) to milli Ampere (mA)
    """
    return amp * 1000



# encoding: utf-8
"""Collection emulator, calling an external program to generate data images.
Written originally for Global Phasing simcal,
but could be made to work with other systems

License:

This file is part of MXCuBE.

MXCuBE is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MXCuBE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with MXCuBE. If not, see <https://www.gnu.org/licenses/>.
"""

import os
import re
import subprocess
from collections import OrderedDict

import f90nml

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.mockup.CollectMockup import CollectMockup
from mxcubecore.TaskUtils import task
from mxcubecore.utils import conversion

__copyright__ = """ Copyright © 2017 - 2019 by Global Phasing Ltd. """
__license__ = "LGPLv3+"
__author__ = "Rasmus H Fogh"


class CollectEmulator(CollectMockup):
    def __init__(self, name):
        CollectMockup.__init__(self, name)

        self._counter = 1

    def _get_simcal_input(self, data_collect_parameters, crystal_data):
        """Get ordered dict with simcal input from available data"""

        # Set up and add crystal data
        result = OrderedDict()
        setup_data = result["setup_list"] = crystal_data
        instrument_data = HWR.beamline.gphl_workflow.instrument_data

        #
        # # update with instrument data
        #
        sweep_count = len(data_collect_parameters["oscillation_sequence"])

        # Setting parameters in order (may not be necessary, but ...)
        # Missing: *mu*
        remap = {
            "beam": "nominal_beam_dir",
            "det_coord_def": "det_org_dist",
            "cone_s_height": "cone_height",
        }
        tags = (
            "lambda_sd",
            "beam",
            "beam_sd_deg",
            "pol_plane_n",
            "pol_frac",
            "d_sensor",
            "min_zeta",
            "det_name",
            "det_x_axis",
            "det_y_axis",
            "det_qx",
            "det_qy",
            "det_nx",
            "det_ny",
            # "det_org_x",
            # "det_org_y",
            "det_coord_def",
        )
        for tag in tags:
            val = instrument_data.get(remap.get(tag, tag))
            if val is not None:
                setup_data[tag] = val

        (
            setup_data["det_org_x"],
            setup_data["det_org_y"],
        ) = HWR.beamline.detector.get_beam_position()

        ll0 = list(HWR.beamline.gphl_workflow.rotation_axes.values())
        setup_data["omega_axis"] = ll0[0]
        setup_data["kappa_axis"] = ll0[1]
        setup_data["phi_axis"] = ll0[2]
        ll0 = list(HWR.beamline.gphl_workflow.translation_axes.values())
        setup_data["trans_x_axis"] = ll0[0]
        setup_data["trans_y_axis"] = ll0[1]
        setup_data["trans_z_axis"] = ll0[2]
        tags = (
            "cone_radius",
            "cone_s_height",
            "beam_stop_radius",
            "beam_stop_s_length",
            "beam_stop_s_distance",
        )
        for tag in tags:
            val = instrument_data.get(remap.get(tag, tag))
            if val is not None:
                setup_data[tag] = val

        # Add/overwrite parameters from emulator configuration
        conv = conversion.convert_string_value
        for key, val in self.config.simcal_parameters.items():
            setup_data[key] = conv(val)

        setup_data["n_vertices"] = 0
        setup_data["n_triangles"] = 0
        setup_data["n_orients"] = 0
        setup_data["n_sweeps"] = sweep_count

        # Add segments
        segments = HWR.beamline.gphl_workflow.detector_segments
        if segments:
            if isinstance(segments, dict):
                segment_count = 1
            else:
                segment_count = len(segments)
            setup_data["n_segments"] = segment_count
            result["segment_list"] = segments

        # Adjustments
        val = instrument_data.get("beam")
        if val:
            setup_data["beam"] = val

        # update with diffractcal data
        # TODO check that this works also for updating segment list
        fp0 = HWR.beamline.gphl_workflow.file_paths.get("diffractcal_file")
        if os.path.isfile(fp0):
            diffractcal_data = f90nml.read(fp0)["sdcp_instrument_list"]
            for tag in setup_data.keys():
                val = diffractcal_data.get(tag)
                if val is not None:
                    setup_data[tag] = val
            ll0 = diffractcal_data["gonio_axis_dirs"]
            setup_data["omega_axis"] = ll0[:3]
            setup_data["kappa_axis"] = ll0[3:6]
            setup_data["phi_axis"] = ll0[6:]

        # get resolution limit and detector distance
        detector_distance = data_collect_parameters.get("detector_distance", 0.0)
        if not detector_distance:
            dd = data_collect_parameters.get("resolution")
            # resolution may not be set - if so you should take the current value
            if dd:
                resolution = dd.get("upper")
                if resolution:
                    self.set_resolution(resolution)
            detector_distance = HWR.beamline.detector.distance.get_value()
        # Add sweeps
        sweeps = []
        compress_data = False
        for osc in data_collect_parameters["oscillation_sequence"]:
            motors = data_collect_parameters["motors"]
            sweep = OrderedDict()

            energy = (
                data_collect_parameters.get("energy") or HWR.beamline.energy.get_value()
            )
            sweep["lambda"] = conversion.HC_OVER_E / energy
            sweep["res_limit"] = setup_data["res_limit_def"]
            sweep["exposure"] = osc["exposure_time"]
            ll0 = list(HWR.beamline.gphl_workflow.translation_axes)
            sweep["trans_xyz"] = list(motors.get(x) or 0.0 for x in ll0)
            sweep["det_coord"] = detector_distance
            # NBNB hardwired for omega scan TODO
            sweep["axis_no"] = 3
            sweep["omega_deg"] = osc["start"]
            # NB kappa and phi are overwritten from the motors dict, if set there
            sweep["kappa_deg"] = osc["kappaStart"]
            sweep["phi_deg"] = osc["phiStart"]
            sweep["step_deg"] = osc["range"]
            sweep["n_frames"] = osc["number_of_images"]
            sweep["image_no"] = osc["start_image_number"]
            # self.make_image_file_template(data_collect_parameters, suffix='cbf')

            # Extract format statement from template,
            # and convert to fortran format
            text_type = conversion.text_type
            template = text_type(data_collect_parameters["fileinfo"]["template"])
            ss0 = re.search("(%[0-9]+d)", template).group(0)
            template = template.replace(ss0, "?" * int(ss0[1:-1]))
            name_template = os.path.join(
                text_type(data_collect_parameters["fileinfo"]["directory"]),
                template,
                # data_collect_parameters['fileinfo']['template']
            )
            # We still use the normal name template for compressed data
            if name_template.endswith(".gz"):
                compress_data = True
                name_template = name_template[:-3]
            sweep["name_template"] = name_template

            # Overwrite kappa and phi from motors - if set
            val = motors.get("kappa")
            if val is not None:
                sweep["kappa_deg"] = val
            val = motors.get("kappa_phi")
            if val is not None:
                sweep["phi_deg"] = val

            # Skipped: spindle_deg=0.0, two_theta_deg=0.0, mu_air=-1, mu_sensor=-1

            sweeps.append(sweep)

        if len(sweeps) == 1:
            # NBNB in current code we can have only one sweep here,
            # but it will work for multiple
            result["sweep_list"] = sweep
        else:
            result["sweep_list"] = sweeps
        return result, compress_data

    @task
    def data_collection_hook(self):
        """Spawns data emulation using gphl simcal"""
        data_model = HWR.beamline.gphl_workflow._queue_entry.get_data_model()
        if data_model.skip_collection:
            return

        data_collect_parameters = self.current_dc_parameters

        if not HWR.beamline.gphl_workflow:
            raise ValueError("Emulator requires GPhL workflow installation")
        gphl_connection = HWR.beamline.gphl_connection
        if not gphl_connection:
            raise ValueError("Emulator requires GPhL connection installation")

        # Get program locations
        simcal_executive = gphl_connection.get_executable("simcal")
        simcal_licence_dir = (
            gphl_connection.get_bdg_licence_dir("simcal")
            or gphl_connection.config.software_paths["GPHL_INSTALLATION"]
        )
        # # Get environmental variables.
        envs = {"autoPROC_home": simcal_licence_dir}
        GPHL_XDS_PATH = gphl_connection.config.software_paths.get("GPHL_XDS_PATH")
        if GPHL_XDS_PATH:
            envs["GPHL_XDS_PATH"] = GPHL_XDS_PATH
        GPHL_CCP4_PATH = gphl_connection.config.software_paths.get("GPHL_CCP4_PATH")
        if GPHL_CCP4_PATH:
            envs["GPHL_CCP4_PATH"] = GPHL_CCP4_PATH
        text_type = conversion.text_type
        for tag, val in self.config.environment_variables.items():
            envs[text_type(tag)] = text_type(val)

        # get crystal data
        crystal_data, hklfile = HWR.beamline.gphl_workflow.get_emulation_crystal_data()

        input_data, compress_data = self._get_simcal_input(
            data_collect_parameters, crystal_data
        )

        # NB outfile is the echo output of the input file;
        # image files templates are set in the input file
        file_info = data_collect_parameters["fileinfo"]
        if not os.path.exists(file_info["directory"]):
            os.makedirs(file_info["directory"])
        infile = os.path.join(
            file_info["directory"], "simcal_in_%s.nml" % self._counter
        )

        f90nml.write(input_data, infile, force=True)
        outfile = os.path.join(
            file_info["directory"], "simcal_out_%s.nml" % self._counter
        )
        logfile = os.path.join(
            file_info["directory"], "simcal_log_%s.txt" % self._counter
        )
        self._counter += 1
        command_list = [
            simcal_executive,
            "--input",
            infile,
            "--output",
            outfile,
            "--hkl",
            hklfile,
        ]

        for tag, val in self.config.simcal_options.items():
            command_list.extend(conversion.command_option(tag, val, prefix="--"))
        self.log.info("Executing command: %s", " ".join(command_list))
        self.log.info("Executing environment: %s", sorted(envs.items()))

        if compress_data:
            command_list.append("--gzip-img")

        fp1 = open(logfile, "w")
        fp2 = subprocess.STDOUT
        # resource.setrlimit(resource.RLIMIT_STACK, (-1,-1))

        def set_ulimit():
            import resource

            resource.setrlimit(
                resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY)
            )

        try:
            self.emit("collectStarted", (None, 1))
            running_process = subprocess.Popen(
                command_list, stdout=fp1, stderr=fp2, env=envs, preexec_fn=set_ulimit
            )
            gphl_connection.collect_emulator_process = running_process

            # # NBNB leaving this in causes simulations to be killed and missing images
            # super(CollectEmulator, self).data_collection_hook()

            self.log.info("Waiting for simcal collection emulation.")
            # NBNB TODO put in time-out, somehow
            return_code = running_process.wait()
        except Exception:
            self.log.error("Error in GPhL collection emulation")
            raise
        finally:
            fp1.close()
        process = gphl_connection.collect_emulator_process
        gphl_connection.collect_emulator_process = None
        if process == "ABORTED":
            self.log.info("Simcal collection emulation aborted")
        elif return_code:
            raise RuntimeError(
                "simcal process terminated with return code %s" % return_code
            )
        else:
            self.log.info("Simcal collection emulation successful")
        self.ready_event.set()
        return



"""Provides standard loggers.

Defines MXCuBE-core standard loggers.

  `hwr_log` for logging to HWR log
  `user_log` for logging to user_level_log log
"""

from logging import getLogger

hwr_log = getLogger("HWR")
user_log = getLogger("user_level_log")



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract class for a Sample View.
Defines methods to handle snapshots, animation and shapes.
"""

__copyright__ = """2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"

import abc
import logging
from typing import (
    Literal,
    Union,
)

from mxcubecore.BaseHardwareObjects import HardwareObject

ShapeState = Literal["HIDDEN", "SAVED", "TMP"]


class AbstractSampleView(HardwareObject):
    """AbstractSampleView Class"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self._focus = None
        self._zoom = None
        self._frontlight = None
        self._backlight = None
        self._shapes = {}
        self.current_centring_procedure = None

    @property
    def camera(self):
        return self.get_object_by_role("camera")

    @abc.abstractmethod
    def get_snapshot(
        self,
        overlay: bool | str = True,
        bw: bool = False,
        return_as_array: bool = False,
    ):
        """Get snappshot(s)
        Args:
            overlay(bool | str): Display shapes and other items on the snapshot
            bw(bool): return grayscale image
            return_as_array(bool): return as np array
        """

    @abc.abstractmethod
    def save_snapshot(
        self, filename, overlay: Union[bool, str] = True, bw: bool = False
    ):
        """Save a snapshot to file.
        Args:
            filename (str): The filename.
            overlay(bool | str): Display shapes and other items on the snapshot.
            bw(bool): Return grayscale image.
        """

    def save_scene_animation(self, filename, duration=1):
        """Take snapshots and create an animation.
        Args:
            filename (str): The filename.
            duration (int): Duration time [s].
        """

    @property
    def shapes(self):
        """Get shapes dict.
        Returns:
            (AbstractShapes): Shapes hardware object.
        """
        return self._shapes

    @property
    def zoom(self):
        """Get zoom object.
        Returns:
            (AbstractZoom): Zoom gardware object.
        """
        return self._zoom

    @property
    def frontlight(self):
        """Get Front light object
        Returns:
            (AbstractLight): Front light hardware object.
        """
        return self._frontlight

    @property
    def backlight(self):
        """Get Back light object.
        Returns:
            (AbstractLight): Back light hardware object.
        """
        return self._backlight

    @abc.abstractmethod
    def start_manual_centring(self, nb_click=3):
        """Starts manual centring procedure"""

    @abc.abstractmethod
    def start_auto_centring(self):
        """Start automatic centring procedure"""

    def move_to_beam(self, x: float, y: float):
        """Move the sample to the x,y coordinates"""

    def cancel_centring(self):
        """Cancels current centring procedure"""
        if self.current_centring_procedure:
            try:
                self.current_centring_procedure.kill(block=True)
            except Exception:
                logging.getLogger("HWR").exception(
                    "Problem aborting the centring method"
                )
            self.current_centring_procedure = None
            self.emit("centringFailed")
            logging.getLogger("HWR").exception("Centring canceled")

    @abc.abstractmethod
    def add_shape(self, shape):
        """Add the shape <shape> to the dictionary of handled shapes.
        Args:
            shape(Shape): Shape to add
        """
        return

    @abc.abstractmethod
    def add_shape_from_mpos(
        self,
        mpos_list,
        screen_cord,
        _type,
        state: ShapeState = "SAVED",
        user_state: ShapeState = "SAVED",
    ):
        """Add a shape of type <t>, with motor positions from mpos_list and
        screen position screen_coord.
        Args:
            mpos_list (list[mpos_list]): List of motor positions
            screen_coord (tuple(x, y): Screen coordinate for shape
            _type (str): Type str for shape, P (Point), L (Line), G (Grid)
            user_state (ShapeState): State of the shape set by the user
        Returns:
            (Shape): Shape of type _type
        """
        return

    @abc.abstractmethod
    def delete_shape(self, sid):
        """Remove the shape with specified id from the list of handled shapes.
        Args:
            sid (str): The id of the shape to remove
        Returns:
            (Shape): The removed shape
        """
        return

    @abc.abstractmethod
    def select_shape(self, sid):
        """Select the shape <shape>.
        Args:
            sid (str): Id of the shape to select.
        """
        return

    @abc.abstractmethod
    def de_select_shape(self, sid):
        """De-select the shape with id <sid>.
        Args:
            sid (str): The id of the shape to de-select.
        """
        return

    @abc.abstractmethod
    def is_selected(self, sid):
        """Check if Shape with specified id is selected.
        Args:
            sid (int): Shape id.
        Returns:
            (Boolean) True if selected, False otherwise.
        """

    @abc.abstractmethod
    def get_selected_shapes(self):
        """Get all selected shapes.
        Returns:
           (list) List of the selected Shapes.
        """
        return

    @abc.abstractmethod
    def de_select_all(self):
        """De-select all shapes."""
        return

    @abc.abstractmethod
    def select_shape_with_cpos(self, cpos):
        """Selects shape with the assocaitaed centred position <cpos>
        Args:
            cpos (CentredPosition): Centred position
        """
        return

    @abc.abstractmethod
    def clear_all(self):
        """
        Clear the shapes, remove all contents.
        """
        return

    @abc.abstractmethod
    def get_shape(self, sid: str):
        """
        Get Shape with id <sid>.

        Args:
            sid (str): id of Shape to retrieve

        Returns:
            (Shape) All the shapes
        """
        return

    @abc.abstractmethod
    def get_grid(self):
        """Get the first of the selected grids, (the one that was selected
        first in a sequence of select operations).
        Returns:
            (dict): The first selected grid as a dictionary.
        """
        return

    @abc.abstractmethod
    def get_points(self):
        """Get all currently handled centred points.
        Returns:
            (list): All points currently handled as list.
        """
        return

    @abc.abstractmethod
    def get_lines(self):
        """Get all the currently handled lines.

        Returns:
            (list): All lines currently handled as list.
        """
        return

    @abc.abstractmethod
    def get_grids(self):
        """Get all currently handled grids.
        Returns:
            (list): All grids currently handled as list.
        """
        return

    @abc.abstractmethod
    def inc_used_for_collection(self, cpos):
        """Increase the counter that keepts on collect made on this shape,
        shape with associated CentredPosition cpos.
        Args:
            cpos (CentredPosition): CentredPosition of shape
        """

    def motor_positions_to_screen(self, positions_dict: dict) -> tuple:
        """Get the motor positions according to the calibration"""

        return ()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
Example of xml config file

<object class="MotorMockup">
  <username>Mock motor</username>
  <actuator_name>mock_motor</actuator_name>
  <!-- for the mockup only -->
  <default_value>500</default_value>
  <velocity>100</velocity>
  <wrap_range>None</wrap_range>
  <default_limits>[-360, 360]</default_limits>
</object>
"""

import time

from mxcubecore.HardwareObjects.abstract.AbstractMotor import AbstractMotor, MotorStates
from mxcubecore.HardwareObjects.mockup.ActuatorMockup import ActuatorMockup

__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"

DEFAULT_VELOCITY = 100
DEFAULT_LIMITS = (-10000, 10000)
DEFAULT_VALUE = 10.124
DEFAULT_WRAP_RANGE = None


class MotorMockup(ActuatorMockup, AbstractMotor):
    """Mock Motor implementation"""

    SPECIFIC_STATES = MotorStates

    def __init__(self, name):
        AbstractMotor.__init__(self, name)
        self._wrap_range = None

    def init(self):
        """Initialisation method"""
        # get username, actuator_name and tolerance
        super().init()

        # local properties
        if not self.get_velocity():
            self.set_velocity(DEFAULT_VELOCITY)
        if None in self.get_limits():
            self.update_limits(DEFAULT_LIMITS)

        try:
            self._wrap_range = float(self.get_property("wrap_range"))
        except (TypeError, ValueError):
            self._wrap_range = DEFAULT_WRAP_RANGE

        if self.default_value is None:
            self.default_value = DEFAULT_VALUE
            self.update_value(self.default_value)

        self.update_state(self.STATES.READY)

    def _move(self, value):
        """Simulated motor movement.
        Args:
            value (float): target position
        Returns:
            (float): The reached position.
        """

        self.update_specific_state(self.SPECIFIC_STATES.MOVING)

        start_pos = self.get_value()
        if value is not None and start_pos is not None:
            delta = abs(value - start_pos)

            direction = -1 if value < self.get_value() else 1

            start_time = time.time()

            while (time.time() - start_time) < (delta / self.get_velocity()):
                time.sleep(0.02)
                val = start_pos + direction * self.get_velocity() * (
                    time.time() - start_time
                )

                val = val if not self._wrap_range else val % self._wrap_range

                self.update_value(val)
        time.sleep(0.02)

        _low, _high = self.get_limits()
        if value == self.default_value:
            self.update_specific_state(self.SPECIFIC_STATES.HOME)
        elif value == _low:
            self.update_specific_state(self.SPECIFIC_STATES.LOWLIMIT)
        elif value == _high:
            self.update_specific_state(self.SPECIFIC_STATES.HIGHLIMIT)
        else:
            self.update_specific_state(None)

        return value

    def is_moving(self):
        return (self.get_state() == self.STATES.BUSY) or (
            self.get_state() == self.SPECIFIC_STATES.MOVING
        )
